from __future__ import division
from __future__ import print_function

import warnings
from collections import defaultdict

from ..utils.utility import _get_sklearn_version

#if _get_sklearn_version() > 20:
from inspect import signature
#else:
#    from sklearn.externals.funcsigs import signature

import abc
import six

import numpy as np
from numpy import percentile
from scipy.special import erf
from sklearn.preprocessing import MinMaxScaler
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.multiclass import check_classification_targets
from .sklearn_base import _pprint


@six.add_metaclass(abc.ABCMeta)
class DetectorB(object):
    """Abstract class for type B anomoly detection algorithms.
    Parameters
    ----------
    contamination : float, optional (default=0.1)
        The proportion of anomolies of the data set,
        Used for defining the threshold on the decision function.
    window: int, optional (default = 20)
        The length of the window to detect the given anomolies 
    Attributes
    ----------
    decision_scores_ : numpy array of shape (n_samples,)
        The outlier scores of the training data.
        The higher, the more abnormal. Outliers tend to have higher
        scores. This value is available once the detector is fitted.
    threshold_ : float
        The threshold is based on ``contamination``. It is the
        ``n_samples * contamination`` most abnormal samples in
        ``decision_scores_``. The threshold is calculated for generating
        binary outlier labels.
    labels_ : int, either 0 or 1
        The binary labels of the training data. 0 stands for inliers
        and 1 for outliers/anomalies. It is generated by applying
        ``threshold_`` on ``decision_scores_``.
    """

    @abc.abstractmethod
    def __init__(self, contamination=0.1, window = 75):

        if not (0. < contamination <= 0.55):
            raise ValueError("contamination must be in (0, 0.55], "
                             "got: %f" % contamination)

        self.contamination = contamination

    # noinspection PyIncorrectDocstring
    @abc.abstractmethod
    def fit(self, X, y=None):
        """Fit detector. y is ignored in unsupervised methods.
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The input samples.
        feature: feature class that maps timeseires to the feature space
        y : Ignored
            Not used, present for API consistency by convention.
        Returns
        -------
        self : object
            Fitted estimator.
        """
        pass

    @abc.abstractmethod
    def decision_function(self, X):
        """Predict raw anomaly scores of X using the unsupervised method
        The anomaly score of an input sample is computed based on image of X
        under the feature map and the models of the detector. For consistency,
         outliers are assigned with higher anomaly scores.
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The input samples. Sparse matrices are accepted only
            if they are supported by the base estimator.
        Returns
        -------
        anomaly_scores : numpy array of shape (n_samples,)
            The anomaly score of the input samples.
        """
        pass

    def predict(self, X):
        """Predict if a particular sample is an outlier or not.
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The input samples.
        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        check_is_fitted(self, ['decision_scores_', 'threshold_', 'labels_'])

        pred_score = self.decision_function(X)
        return (pred_score > self.threshold_).astype('int').ravel()

    def predict_proba(self, method='linear'):
        """Predict the probability of a sample being outlier. Two approaches
        are possible:
        1. simply use Min-max conversion to linearly transform the outlier
           scores into the range of [0,1]. The model must be
           fitted first.
        2. use unifying scores, see :cite:`kriegel2011interpreting`.
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The input samples.
        method : str, optional (default='linear')
            probability conversion method. It must be one of
            'linear' or 'unify'.
        Returns
        -------
        outlier_probability : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. Return the outlier probability, ranging
            in [0,1].
        """

        check_is_fitted(self, ['decision_scores_', 'threshold_', 'labels_'])
        train_scores = self.train_decision_scores_

        test_scores = self.decision_scores_

        probs = np.zeros([self.n_train_, int(self._classes)])
        if method == 'linear':
            scaler = MinMaxScaler().fit(train_scores.reshape(-1, 1))
            probs[:, 1] = scaler.transform(
                test_scores.reshape(-1, 1)).ravel().clip(0, 1)
            probs[:, 0] = 1 - probs[:, 1]
            return probs

        elif method == 'unify':
            # turn output into probability
            pre_erf_score = (test_scores - self._mu) / (
                    self._sigma * np.sqrt(2))
            erf_score = erf(pre_erf_score)
            probs[:, 1] = erf_score.clip(0, 1).ravel()
            probs[:, 0] = 1 - probs[:, 1]
            return probs
        else:
            raise ValueError(method,
                             'is not a valid probability conversion method')

    def _predict_rank(self, normalized=False):
        """Predict the outlyingness rank of a sample by a fitted model. The
        method is for outlier detector score combination.
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The input samples.
        normalized : bool, optional (default=False)
            If set to True, all ranks are normalized to [0,1].
        Returns
        -------
        ranks : array, shape (n_samples,)
            Outlying rank of a sample according to the training data.
        """

        check_is_fitted(self, ['decision_scores_'])

        train_scores = self.train_decision_scores_

        test_scores = self.decision_scores_

        sorted_train_scores = np.sort(train_scores)
        ranks = np.searchsorted(sorted_train_scores, test_scores)

        if normalized:
            # return normalized ranks
            ranks = ranks / ranks.max()
        return ranks

 

    def _set_n_classes(self, y):
        """Set the number of classes if `y` is presented, which is not
        expected. It could be useful for multi-class outlier detection.
        Parameters
        ----------
        y : numpy array of shape (n_samples,)
            Ground truth.
        Returns
        -------
        self
        """

        self._classes = 2  # default as binary classification
        if y is not None:
            check_classification_targets(y)
            self._classes = len(np.unique(y))
            warnings.warn(
                "y should not be presented in unsupervised learning.")
        return self

    
    def _process_decision_scores(self):
        """Internal function to calculate key attributes:
        - threshold_: used to decide the binary label
        - labels_: binary labels of training data
        Returns
        -------
        self
        """

        self.threshold_ = percentile(self.decision_scores_,
                                     100 * (1 - self.contamination))
        self.labels_ = (self.decision_scores_ > self.threshold_).astype(
            'int').ravel()

        # calculate for predict_proba()

        self._mu = np.mean(self.decision_scores_)
        self._sigma = np.std(self.decision_scores_)

        return self

    def _set_threshold(self, X_train):
        """Internal function to calculate key attributes:
        - threshold_: used to decide the binary label
        - labels_: binary labels of training data
        - measure: the measure class to define the set
        Returns
        -------
        self
        """
        self.fit(X_train)
    

        self.train_decision_scores_ = self.decision_scores_
    
        self.threshold_ = percentile(self.decision_scores_,
                                     100 * (1 - self.contamination))
        self.labels_ = (self.decision_scores_ > self.threshold_).astype(
            'int').ravel()

        # calculate for predict_proba()

        self._mu = np.mean(self.decision_scores_)
        self._sigma = np.std(self.decision_scores_)

        return self
    # noinspection PyMethodParameters
    def _get_param_names(cls):
        # noinspection PyPep8
        """Get parameter names for the estimator
        See http://scikit-learn.org/stable/modules/generated/sklearn.base.BaseEstimator.html
        and sklearn/base.py for more information.
        """

        # fetch the constructor or the original constructor before
        # deprecation wrapping if any
        init = getattr(cls.__init__, 'deprecated_original', cls.__init__)
        if init is object.__init__:
            # No explicit constructor to introspect
            return []

        # introspect the constructor arguments to find the model parameters
        # to represent
        init_signature = signature(init)
        # Consider the constructor parameters excluding 'self'
        parameters = [p for p in init_signature.parameters.values()
                      if p.name != 'self' and p.kind != p.VAR_KEYWORD]
        for p in parameters:
            if p.kind == p.VAR_POSITIONAL:
                raise RuntimeError("scikit-learn estimators should always "
                                   "specify their parameters in the signature"
                                   " of their __init__ (no varargs)."
                                   " %s with constructor %s doesn't "
                                   " follow this convention."
                                   % (cls, init_signature))
        # Extract and sort argument names excluding 'self'
        return sorted([p.name for p in parameters])

    # noinspection PyPep8
    def get_params(self, deep=True):
        """Get parameters for this estimator.
        See http://scikit-learn.org/stable/modules/generated/sklearn.base.BaseEstimator.html
        and sklearn/base.py for more information.
        Parameters
        ----------
        deep : bool, optional (default=True)
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.
        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        """

        out = dict()
        for key in self._get_param_names():
            # We need deprecation warnings to always be on in order to
            # catch deprecated param values.
            # This is set in utils/__init__.py but it gets overwritten
            # when running under python3 somehow.
            warnings.simplefilter("always", DeprecationWarning)
            try:
                with warnings.catch_warnings(record=True) as w:
                    value = getattr(self, key, None)
                if len(w) and w[0].category == DeprecationWarning:
                    # if the parameter is deprecated, don't show it
                    continue
            finally:
                warnings.filters.pop(0)

            # XXX: should we rather test if instance of estimator?
            if deep and hasattr(value, 'get_params'):
                deep_items = value.get_params().items()
                out.update((key + '__' + k, val) for k, val in deep_items)
            out[key] = value
        return out

    def set_params(self, **params):
        # noinspection PyPep8
        """Set the parameters of this estimator.
        The method works on simple estimators as well as on nested objects
        (such as pipelines). The latter have parameters of the form
        ``<component>__<parameter>`` so that it's possible to update each
        component of a nested object.
        See http://scikit-learn.org/stable/modules/generated/sklearn.base.BaseEstimator.html
        and sklearn/base.py for more information.
        Returns
        -------
        self : object
        """

        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        valid_params = self.get_params(deep=True)

        nested_params = defaultdict(dict)  # grouped by prefix
        for key, value in params.items():
            key, delim, sub_key = key.partition('__')
            if key not in valid_params:
                raise ValueError('Invalid parameter %s for estimator %s. '
                                 'Check the list of available parameters '
                                 'with `estimator.get_params().keys()`.' %
                                 (key, self))

            if delim:
                nested_params[key][sub_key] = value
            else:
                setattr(self, key, value)

        for key, sub_params in nested_params.items():
            valid_params[key].set_params(**sub_params)

        return self

    def __repr__(self):
        # noinspection PyPep8
        """
        See http://scikit-learn.org/stable/modules/generated/sklearn.base.BaseEstimator.html
        and sklearn/base.py for more information.
        """

        class_name = self.__class__.__name__
        return '%s(%s)' % (class_name, _pprint(self.get_params(deep=False),
                                               offset=len(class_name), ),)


"""Local Outlier Factor (LOF). Implemented on scikit-learn library.
"""
# Author: Yue Zhao <zhaoy@cmu.edu>


from __future__ import division
from __future__ import print_function

from sklearn.neighbors import LocalOutlierFactor
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.validation import check_array

from .detectorB import DetectorB
from ..utils.utility import invert_order
from ..utils.utility import _get_sklearn_version


class LOF(DetectorB):
    """
    Wrapper of scikit-learn LOF Class with more functionalities.

    Parameters
    ----------
    n_neighbors : int, optional (default=20)
        Number of neighbors to use by default for `kneighbors` queries.
        If n_neighbors is larger than the number of samples provided,
        all samples will be used.
    algorithm : {'auto', 'ball_tree', 'kd_tree', 'brute'}, optional
        Algorithm used to compute the nearest neighbors:

        - 'ball_tree' will use BallTree
        - 'kd_tree' will use KDTree
        - 'brute' will use a brute-force search.
        - 'auto' will attempt to decide the most appropriate algorithm
          based on the values passed to :meth:`fit` method.

        Note: fitting on sparse input will override the setting of
        this parameter, using brute force.
    leaf_size : int, optional (default=30)
        Leaf size passed to `BallTree` or `KDTree`. This can
        affect the speed of the construction and query, as well as the memory
        required to store the tree. The optimal value depends on the
        nature of the problem.
    metric : string or callable, default 'minkowski'
        metric used for the distance computation. Any metric from scikit-learn
        or scipy.spatial.distance can be used.
        If 'precomputed', the training input X is expected to be a distance
        matrix.
        If metric is a callable function, it is called on each
        pair of instances (rows) and the resulting value recorded. The callable
        should take two arrays as input and return one value indicating the
        distance between them. This works for Scipy's metrics, but is less
        efficient than passing the metric name as a string.
        Valid values for metric are:

        - from scikit-learn: ['cityblock', 'cosine', 'euclidean', 'l1', 'l2',
          'manhattan']
        - from scipy.spatial.distance: ['braycurtis', 'canberra', 'chebyshev',
          'correlation', 'dice', 'hamming', 'jaccard', 'kulsinski',
          'mahalanobis', 'matching', 'minkowski', 'rogerstanimoto',
          'russellrao', 'seuclidean', 'sokalmichener', 'sokalsneath',
          'sqeuclidean', 'yule']

    p : integer, optional (default = 2)
        Parameter for the Minkowski metric from
        sklearn.metrics.pairwise.pairwise_distances. When p = 1, this is
        equivalent to using manhattan_distance (l1), and euclidean_distance
        (l2) for p = 2. For arbitrary p, minkowski_distance (l_p) is used.
        See http://scikit-learn.org/stable/modules/generated/sklearn.metrics.pairwise.pairwise_distances
    metric_params : dict, optional (default = None)
        Additional keyword arguments for the metric function.
    contamination : float in (0., 0.5), optional (default=0.1)
        The amount of contamination of the data set, i.e. the proportion
        of outliers in the data set. When fitting this is used to define the
        threshold on the decision function.
    n_jobs : int, optional (default = 1)
        The number of parallel jobs to run for neighbors search.
        If ``-1``, then the number of jobs is set to the number of CPU cores.
        Affects only kneighbors and kneighbors_graph methods.
    
    Attributes
    ----------
    decision_scores_ : numpy array of shape (n_samples,)
        The outlier scores of the training data.
        The higher, the more abnormal. Outliers tend to have higher
        scores. This value is available once the detector is
        fitted.
    labels_ : int, either 0 or 1
        The binary labels of the training data. 0 stands for inliers
        and 1 for outliers/anomalies. It is generated by applying
        ``threshold_`` on ``decision_scores_``.
    threshold_ : float
        The threshold is based on ``contamination``. It is the
        ``n_samples * contamination`` most abnormal samples in
        ``decision_scores_``. The threshold is calculated for generating
        binary outlier labels.
    """

    def __init__(self, n_neighbors=20, algorithm='auto', leaf_size=30,
                 metric='minkowski', p=2, metric_params=None,
                 contamination=0.1, n_jobs=1):
        super(LOF, self).__init__(contamination=contamination)
        self.n_neighbors = n_neighbors
        self.algorithm = algorithm
        self.leaf_size = leaf_size
        self.metric = metric
        self.p = p
        self.metric_params = metric_params
        self.n_jobs = n_jobs
        self.model_name = 'LOF'

    
    def fit(self, X, y=None):
        """Fit detector. y is ignored in unsupervised methods.
        
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The input samples (time series length). n_features corresponds to the subsequence length.
        y : Ignored
            Not used, present for API consistency by convention.
        
        Returns
        -------
        self : object
            Fitted estimator.
        """
        
        # validate inputs X and y (optional)
        
        try: 
            X = check_array(X)
        except:
            X = X.reshape(-1, 1)
        
        X = check_array(X)
        

        self.detector_ = LocalOutlierFactor(n_neighbors=self.n_neighbors,
                                            algorithm=self.algorithm,
                                            leaf_size=self.leaf_size,
                                            metric=self.metric,
                                            p=self.p,
                                            metric_params=self.metric_params,
                                            contamination=self.contamination,
                                            n_jobs=self.n_jobs)
        self.detector_.fit(X=X, y=y)

        # Invert decision_scores_. Outliers comes with higher outlier scores
        self.decision_scores_ = -self.detector_.negative_outlier_factor_
        # self._process_decision_scores()
        return self

    def decision_function(self, X):
        """
        Predict raw anomaly score of X using the fitted detector.
        The anomaly score of an input sample is computed based on different
        detector algorithms. For consistency, outliers are assigned with
        larger anomaly scores.

        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The training input samples. Sparse matrices are accepted only
            if they are supported by the base estimator.
        
        Returns
        -------
        anomaly_scores : numpy array of shape (n_samples,)
            The anomaly score of the input samples.
        """

        check_is_fitted(self, ['decision_scores_', 'threshold_', 'labels_'])

        # Invert outlier scores. Outliers comes with higher outlier scores
        # noinspection PyProtectedMember
        if _get_sklearn_version() > 19:
            return invert_order(self.detector_._score_samples(X))
        else:
            return invert_order(self.detector_._decision_function(X))

    @property
    def n_neighbors_(self):
        """The actual number of neighbors used for kneighbors queries.
        Decorator for scikit-learn LOF attributes.

        :meta private:
        """
        return self.detector_.n_neighbors_


# Author: Yuhao Kang <yuhaok@uchicago.edu>

import numpy as np
import networkx as nx 
import math
from .sbd_distance import _sbd


#   SBD distance matrix
def construct_sbd_matrix(X_train, y_train):
    n_indexed = len(y_train)
    X_train_distance = np.zeros([n_indexed,n_indexed])
    
    for i in range(n_indexed):
        for j in range(i+1,n_indexed):
            X_train_distance[i][j] = max(_sbd(X_train[i],X_train[j])[0], 0)
            X_train_distance[j][i] = X_train_distance[i][j]
    
    np.fill_diagonal(X_train_distance, 1e4)

    closest = np.argmin(X_train_distance, axis=0)
    y_pred = y_train[closest]

    n_label = len(np.unique(y_train))

    confusionMatrix = np.zeros((n_label,n_label))
    
            
    for i in range(n_label):    # real label is i+1
        for j in range(n_label):  # predicted label is j+1
            tmp1 = y_train==(i+1)
            tmp2 = y_pred==(j+1)
            confusionMatrix[i][j] = np.sum(tmp1*tmp2)/np.sum(tmp1)
            
    adjacencyMatrix = (confusionMatrix + confusionMatrix.T)/2 + 1e-3
    # np.fill_diagonal(adjacencyMatrix,0)
    return adjacencyMatrix


#   clustering based on max-spanning-tree
def clustering_mst(adjacencyMatrix):
    G = nx.from_numpy_matrix(adjacencyMatrix, parallel_edges=False)
    T = nx.maximum_spanning_tree(G,weight='weight')
        
    n_label = len(adjacencyMatrix)
    labelList = [-1]*n_label
    labelList[0] = 0
    
    # BFS
    visited = [0]
    queue = [0]
    while queue:
        s = queue.pop(0)
        for neighbor in T.neighbors(s):
            if neighbor not in visited:
                visited.append(neighbor)
                queue.append(neighbor)
                labelList[neighbor] = 1 - labelList[s]
                       
    if len(np.unique(labelList))!=2:
        print('Class Separation Error')
        return
        
    labelList = np.array(labelList)
    return labelList


#   construct anomaly based on 2 clusters
def construct_anomaly(X_train, y_train, clusterIndex, K=3, anomalyRatio=0.05, seed=0, return_statistics=False):
    np.random.seed(seed)
    
    # K is num_anomaly, get num of normal
    N_normal_seg = int(K/anomalyRatio)
    
    ano_labels = np.where(clusterIndex==1)[0]+1  # list of ano labels, add 1 since label index start from 1
    nor_labels = np.where(clusterIndex==0)[0]+1  # list of normal labels
    
    # ano_index in y_train
    ano_index = [i for i in range(len(y_train)) if y_train[i] in ano_labels]
    K = min(K, len(ano_index))
    # selected K ano in y_train
    ano_index_selected = np.random.choice(ano_index, size=K, replace=False)
    
    # corresponding K real label 
    ano_label_selected = y_train[ano_index_selected]
    
    # most frequent selected label
    ano_label_mostfreq = max(ano_label_selected.tolist(), key = ano_label_selected.tolist().count)
    freq_ano_label_mostfreq = sum(ano_label_selected==ano_label_mostfreq)
    
    # min freq for each normal label
    freq_nor_label_min = 20 * freq_ano_label_mostfreq  
    # number of normal label will be used
    num_nor_label = min(math.ceil(N_normal_seg/freq_nor_label_min), len(nor_labels))  
    # real freq for each normal label, here the cardinality of each normal label is the same
    freq_nor_label = math.ceil(N_normal_seg/num_nor_label) 
    
    # selected normal labels
    nor_labels_selected = np.random.choice(nor_labels, size=num_nor_label, replace=False)
    # selected index in y_train
    nor_index_selected = []
    for item in nor_labels_selected:
        tmp = np.where(y_train==item)[0]
        nor_index_selected_item = np.random.choice(tmp, size=freq_nor_label, replace=True)
        nor_index_selected.append(nor_index_selected_item)
    nor_index_selected = np.array(nor_index_selected).ravel()
    
    index_combined = np.concatenate((ano_index_selected, nor_index_selected))
    y_new = np.array([1]*len(ano_index_selected) + [0]*len(nor_index_selected))
    
    order = np.arange(len(index_combined))
    np.random.shuffle(order)
    
    X_new = X_train[index_combined[order]].ravel()
    y_new = y_new[order]
    y_new = np.repeat(y_new, len(X_train[0]))
    
    
    if return_statistics:
        # RC 
        X_nor = X_train[nor_index_selected]
        X_ano = X_train[ano_index_selected]
        X_tot = np.concatenate((X_nor, X_ano), axis=0)
        
        n = len(X_tot)
        sbd_matrix = np.zeros([n,n])
        for i in range(n):
            for j in range(i+1,n):
                sbd_matrix[i,j] = max(_sbd(X_tot[i], X_tot[j])[0],0)
                sbd_matrix[j,i] = sbd_matrix[i,j]
        np.fill_diagonal(sbd_matrix, 5)
        
        min_ = np.min(sbd_matrix, 0)
        mean_ = (np.sum(sbd_matrix,0)-5)/(len(sbd_matrix)-1)
        cr = np.mean(mean_)/np.mean(min_)
    
        # NC
        nor_dist_array = sbd_matrix[:len(X_nor), :len(X_nor)].ravel()
        nor_dist_array = nor_dist_array[nor_dist_array<2]
        avg_nor = np.mean(nor_dist_array)
        
        ano_dist_array = sbd_matrix[len(X_nor):n, len(X_nor):n].ravel()
        ano_dist_array = ano_dist_array[ano_dist_array<2]
        avg_ano = np.mean(ano_dist_array)
        
        nc = np.sqrt(avg_nor/avg_ano)
        
        # NA
        if num_nor_label==1:
            na = 1
        else:
            # split index according to label
            cluster_index_basedon_label_ano = []
            for i in np.unique(ano_label_selected):
                cluster_index_basedon_label_ano.append(ano_index_selected[np.where(ano_label_selected==i)])
            
            cluster_index_basedon_label_nor = []
            for i in np.unique(nor_labels_selected):
                cluster_index_basedon_label_nor.append(nor_index_selected[np.where(nor_labels_selected==i)])
                
                
            # get centroids for nor and ano
            centroids_ano = []
            for i in cluster_index_basedon_label_ano:
                centroids_ano.append(np.mean(X_train[i], axis=0))
                
            centroids_nor = []
            for i in cluster_index_basedon_label_nor:
                centroids_nor.append(np.mean(X_train[i], axis=0))
                
            # dist between centroids of nor and ano
            dist_nor_ano = []
            for i in centroids_ano:
                for j in centroids_nor:
                    dist_nor_ano.append(max(_sbd(i,j)[0],0))
                
            # dist between centroids of nor and nor
            dist_nor_nor = []
            for i in range(len(centroids_nor)):
                for j in range(i+1,len(centroids_nor)):
                    dist_nor_nor.append(max(_sbd(centroids_nor[i],centroids_nor[j])[0],0))
                    
            na = np.min(dist_nor_ano)/np.mean(dist_nor_nor)
        
        return X_new, y_new, len(X_train[0]), round(cr,2), round(nc,2), round(na,2)
    
    return X_new, y_new




import time
import os
import math
import pickle
import sys
from tqdm import tqdm
import pandas as pd


import numpy as np
from numpy.random import randint
from numpy.linalg import norm, eigh
from numpy.fft import fft, ifft

from tslearn.clustering import KShape
from tslearn.metrics import cdist_normalized_cc, y_shifted_sbd_vec
from tslearn.utils import to_time_series_dataset,to_time_series

import stumpy



class SAND():
	"""
	Online and offline implementation of SAND
	
	Parameters
	----------
	subsequence_length : int
		Subsequence length to analyze.
	pattern_length : int (greater than subsequence_length)
		Length of the subsequences in Theta.
	k : int (greater than 1), (default=6)
		Number of subsequences in Theta.

	Attributes
	----------
	decision_scores_ : numpy array of shape (n_samples - subsequence_length,)
		The anomaly score.
		The higher, the more abnormal. Anomalies tend to have higher
		scores. This value is available once the detector is
		fitted.
	"""
	def __init__(self,pattern_length,subsequence_length,k=6):
		
		# Configuration parameter
		self.current_time = 0
		self.mean = -1
		self.std = -1

		# algorithm parameter
		self.k = k
		self.subsequence_length = subsequence_length
		self.pattern_length = pattern_length
		
		# real time evolving storage
		self.clusters = []
		self.new_clusters_dist = []
		self.nm_current_weight = []
		self.S = []
		self.clusters_subseqs = []
	
	
	
	def fit(self,X, y=None,online=False,alpha=None,init_length=None,batch_size=None,overlaping_rate=10,verbose=False):
		"""Fit detector. y is ignored in unsupervised methods.
		
		Parameters
		----------

		X : numpy array of shape (n_samples, )
			The input samples.
		y : Ignored
			Not used, present for API consistency by convention.
		online : Boolean, Compute the analysis online or offline

			- if Online, run per batch the model update and the computation of the score (requires to set alpha, init_length, and batch_size)
			- if Offline, run the model for one unique batch
		
		alpha : float ([0,1])
			Update rate (used in Online mode only)
		init_length : int (greater than subsequence_length)
			Length of the initial batch (used in Online mode only).
		batch_size : int (greater than subsequence_length) 
			Length of the batches (used in Online mode only).
		overlaping_rate : int (greater than 1)
			Take subsequence every 'overlaping_rate' points
			
				- Change it to 1 for completely overlapping subsequences 
				- Change it to 'subsequence_length' for non-overlapping subsequences 
				- Change it to 'subsequence_length//4' for non-trivial matching subsequences 


		Returns
		-------
		self : object
			Fitted estimator.
		"""
		self.overlaping_rate = overlaping_rate
		self.ts = list(X)
		self.decision_scores_ = []

		if online:
			if (alpha is None) or (init_length is None) or (batch_size is None):
				print("You must specify a value for alpha, init_length, and batch_size")
				return None
			
			self.alpha = alpha
			self.init_length = init_length
			self.batch_size = batch_size

			if verbose:
				print(self.current_time,end='-->')
			
			self._initialize()
			self._set_normal_model()
			self.decision_scores_ = self._run(self.ts[:min(len(self.ts),self.current_time)])

			while self.current_time < len(self.ts)-self.subsequence_length:

				if verbose:
					print(self.current_time,end='-->')

				self._run_next_batch()
				self._set_normal_model()
				if self.current_time < len(self.ts)-self.subsequence_length:
					self.decision_scores_ += self._run(self.ts[self.current_time-self.batch_size:min(len(self.ts),self.current_time)])
				else:
					self.decision_scores_ += self._run(self.ts[self.current_time-self.batch_size:])




			if verbose:
				print("[STOP]: score length {}".format(len(self.decision_scores_)))


		else:
			self.init_length = len(X)
			self.alpha = 0.5
			self.batch_size = 0
			
			self._initialize()
			self._set_normal_model()
			self.decision_scores_ = self._run(self.ts)

		self.decision_scores_ = np.array(self.decision_scores_)


	# Computation of the anomaly score
	def _run(self,ts):
		all_join = []
		
		all_activated_weighted = []
		if len(self.nm_current_weight) != len(self.weights):
			self.nm_current_weight = self.nm_current_weight + self.weights[len(self.nm_current_weight):]

		for index_name in range(len(self.clusters)):
			if self.nm_current_weight[index_name]> 0:
				join = stumpy.stump(ts,self.pattern_length,self.clusters[index_name][0],ignore_trivial = False)[:,0]
				join = np.array(join)
				join = np.nan_to_num(join)
				all_join.append(join)

		join = [0]*len(all_join[0])
		
		for sub_join,scores_sub_join,scores_sub_join_old,t_decay in zip(all_join,self.weights,self.nm_current_weight,self.time_decay):
			new_w = float(scores_sub_join)/float(1+max(0,t_decay-self.batch_size))
			update_w = float(1-self.alpha)*float(scores_sub_join_old) + float(self.alpha)*float(new_w)

			join = [float(j) + float(sub_j)*update_w for j,sub_j in zip(list(join),list(sub_join))]
			all_activated_weighted.append(update_w)
		
		join = join + [join[-1]]*(self.pattern_length-1)
		join = np.array(join)/np.sum(all_activated_weighted)
		join = self._running_mean(join,self.pattern_length)
		join = [join[0]]*(self.pattern_length-1) + list(join)
		
		self.nm_current_weight = all_activated_weighted
		if self.mean == -1:
			self.mean = np.mean(join)
			self.std = np.std(join)
		else:
			self.mean = (1-self.alpha)*self.mean + self.alpha*np.mean(join)
			self.std = (1-self.alpha)*self.std + self.alpha*np.std(join)

		join = (np.array(join) - self.mean)/self.std
		


		return list(np.nan_to_num(join))


	# MAIN METHODS:
	# - Initialization
	# - Theta update for next batch
	# - Score computaiton

	# Initialization of the model
	def _initialize(self):
		
		cluster_subseqs,clusters = self._kshape_subsequence(initialization=True)

		all_mean_dist = []
		for i,(cluster,cluster_subseq) in enumerate(zip(clusters,cluster_subseqs)):
			self._set_initial_S(cluster_subseq,i,cluster[0])
			all_mean_dist.append(self._compute_mean_dist(cluster[0],cluster[1]))

		self.clusters = clusters
		self.new_clusters_dist = all_mean_dist
		self.current_time = self.init_length




	# Model update for next batch
	def _run_next_batch(self):
		
		# Run K-Shape algorithm on the subsequences of the current batch
		cluster_subseqs,clusters = self._kshape_subsequence(initialization=False)

		#self.new_clusters_subseqs = cluster_subseqs
		self.new_clusters_to_merge = clusters

		to_add = [[] for i in range(len(self.clusters))]
		new_c = []
		
		# Finding the clusters that match exisiting clusters
		# - Storing in to_add all the clusters that have to be merged with the existing clusters
		# - Storing in new_c tyhe new clusters to be added.
		for cluster,cluster_subseq in zip(clusters,cluster_subseqs):
			min_dist = np.Inf
			tmp_index = -1
			for index_o,origin_cluster in enumerate(self.clusters):
				new_dist = self._sbd(origin_cluster[0],cluster[0])[0]
				if min_dist > new_dist:
					min_dist = new_dist
					tmp_index = index_o
			if tmp_index != -1: 
				if min_dist < self.new_clusters_dist[tmp_index]:
					to_add[tmp_index].append((cluster,cluster_subseq))
				else:
					new_c.append((cluster,cluster_subseq))
		
		self.to_add = to_add
		self.new_c = new_c
		
		new_clusters = []
		all_mean_dist = []
		# Merging existing clusters with new clusters
		for i,(cur_c,t_a) in enumerate(zip(self.clusters,to_add)): 
			# Check if new subsequences to add
			if len(t_a) > 0:
				all_index = cur_c[1]
				all_sub_to_add = []
				for t_a_s in t_a:
					all_index += t_a_s[0][1]
					all_sub_to_add +=  t_a_s[1]

				# Updating the centroid shape
				new_centroid,_ = self._extract_shape_stream(all_sub_to_add,i,cur_c[0],initial=False)
				new_clusters.append((self._clean_cluster_tslearn(new_centroid),all_index))

				# Updating the intra cluster distance
				dist_to_add = self._compute_mean_dist(cur_c[0],all_index)
				ratio = float(len(cur_c[1]))/float(len(cur_c[1]) + len(all_index))
				all_mean_dist.append( (ratio) * self.new_clusters_dist[i] + (1.0 - ratio) * dist_to_add )

			# If no new subsequences to add, copy the old cluster
			else:
				new_clusters.append(cur_c)
				all_mean_dist.append(self.new_clusters_dist[i])
		
		# Adding new clusters
		for i,t_a in enumerate(new_c):
			self._set_initial_S(t_a[1],len(self.clusters) + i,t_a[0][0])
			new_clusters.append((t_a[0][0],t_a[0][1]))
			all_mean_dist.append(self._compute_mean_dist(t_a[0][0],t_a[0][1]))
		

		self.clusters = new_clusters
		self.new_clusters_dist = all_mean_dist
		self.current_time = self.current_time + self.batch_size

	
	# SBD distance
	def _sbd(self,x, y):
		ncc = self._ncc_c(x, y)
		idx = ncc.argmax()
		dist = 1 - ncc[idx]
		return dist, None

	# Core clustering computation unit
	def _kshape_subsequence(self,initialization=True):
		all_subsequences = []
		idxs = []
		
		if initialization:
			nb_subsequence = self.init_length
		else:
			nb_subsequence = self.batch_size

		for i in range(self.current_time,min(self.current_time + nb_subsequence,len(self.ts)-self.subsequence_length),self.overlaping_rate):
			all_subsequences.append(self.ts[i:i+self.subsequence_length])
			idxs.append(i)
		
		ks = KShape(n_clusters=self.k,verbose=False)
		list_label = ks.fit_predict(np.array(all_subsequences))
		

		cluster_subseq = [[] for i in range(self.k)]
		cluster_idx = [[] for i in range(self.k)]
		for lbl,idx in zip(list_label,idxs):
			cluster_idx[lbl].append(idx)
			cluster_subseq[lbl].append(self.ts[idx:idx+self.subsequence_length])
		
		# safety check
		new_cluster_subseq = []
		clusters = []

		for i in range(self.k):
			if len(cluster_subseq[i]) > 0:
				new_cluster_subseq.append(cluster_subseq[i])
				clusters.append((self._clean_cluster_tslearn(ks.cluster_centers_[i]),cluster_idx[i]))
		return new_cluster_subseq,clusters




	# Model elements update
	def _set_normal_model(self):
		Frequency = []
		Centrality = []
		Time_decay = []
		for i,nm in enumerate(self.clusters):
			Frequency.append(float(len(nm[1])))
			Time_decay.append(float(self.current_time)-float(nm[1][-1]))
			dist_nms = 0
			for j,nm_t in enumerate(self.clusters):
				if j != i:
					dist_nms += self._sbd(nm[0],nm_t[0])[0]
			Centrality.append(dist_nms)
			
		Frequency = list((np.array(Frequency) - min(Frequency))/(max(Frequency) - min(Frequency)+0.0000001)+1)
		Centrality = list((np.array(Centrality) - min(Centrality))/(max(Centrality) - min(Centrality)+0.0000001)+1)
		
		weights = []
		for f,c,t in zip(Frequency,Centrality,Time_decay):
			weights.append(float(f)**2/float(c))
		
		self.weights = weights
		self.time_decay = Time_decay
		
	# Setting in memory the matrix S
	def _set_initial_S(self,X,idx,cluster_centers):
		X = to_time_series_dataset(X)
		cluster_centers = to_time_series(cluster_centers)
		sz = X.shape[1]
		Xp = y_shifted_sbd_vec(cluster_centers, X,
						norm_ref=-1,
						norms_dataset=np.linalg.norm(X, axis=(1, 2)))
		S = np.dot(Xp[:, :, 0].T, Xp[:, :, 0])
		self.S.append(S)

	# Computation of the updated centroid
	def _extract_shape_stream(self,X,idx,cluster_centers,initial=True):
		X = to_time_series_dataset(X)
		cluster_centers = to_time_series(cluster_centers)
		sz = X.shape[1]
		Xp = y_shifted_sbd_vec(cluster_centers, X,
						norm_ref=-1,
						norms_dataset=np.linalg.norm(X, axis=(1, 2)))
		S = np.dot(Xp[:, :, 0].T, Xp[:, :, 0])

		if not initial:    
			S = S + self.S[idx]
		self.S[idx] = S
		Q = np.eye(sz) - np.ones((sz, sz)) / sz
		M = np.dot(Q.T, np.dot(S, Q))
		_, vec = np.linalg.eigh(M)
		mu_k = vec[:, -1].reshape((sz, 1))
		dist_plus_mu = np.sum(np.linalg.norm(Xp - mu_k, axis=(1, 2)))
		dist_minus_mu = np.sum(np.linalg.norm(Xp + mu_k, axis=(1, 2)))
		if dist_minus_mu < dist_plus_mu:
			mu_k *= -1

		return self._zscore(mu_k, ddof=1),S

	# Reset value of a cluster
	def _clean_cluster_tslearn(self,cluster):
		return np.array([val[0] for val in cluster])
	
	# Compute mean distance of a element in a cluster
	def _compute_mean_dist(self,cluster,all_index):
		dist_all = []
		for i in all_index:
			dist_all.append(self._sbd(self.ts[i:i+self.subsequence_length],cluster)[0])
		return np.mean(dist_all)

	def _running_mean(self,x,N):
		return (np.cumsum(np.insert(x,0,0))[N:] - np.cumsum(np.insert(x,0,0))[:-N])/N

	def _ncc_c(self,x, y):
		den = np.array(norm(x) * norm(y))
		den[den == 0] = np.Inf

		x_len = len(x)
		fft_size = 1 << (2*x_len-1).bit_length()
		cc = ifft(fft(x, fft_size) * np.conj(fft(y, fft_size)))
		cc = np.concatenate((cc[-(x_len-1):], cc[:x_len]))
		return np.real(cc) / den

	def _zscore(self,a, axis=0, ddof=0):
		a = np.asanyarray(a)
		mns = a.mean(axis=axis)
		sstd = a.std(axis=axis, ddof=ddof)
		if axis and mns.ndim < a.ndim:
			res = ((a - np.expand_dims(mns, axis=axis)) /
				   np.expand_dims(sstd, axis=axis))
		else:
			res = (a - mns) / sstd
		return np.nan_to_num(res)
		




"""DAMP.
"""


import numpy as np
import stumpy as st
import pandas as pd


class DAMP():
    """
    Implementation of the DAMP algorithm

    Parameters
    ----------
    m : int,
        target subsequence length.
    sp_index : int
        Need to be strictly greater than m. 
    x_lag : int
        default value None and set to ``2**int(np.ceil(np.log2( 8*self.m )))``
    
    Attributes
    ----------
    decision_scores_ : numpy array of shape (n_samples - m,)
        The anomaly score.
        The higher, the more abnormal. Anomalies tend to have higher
        scores. This value is available once the detector is
        fitted.
    """

    def __init__(self,m,sp_index,x_lag = None):
        
        self.m = m
        self.sp_index = sp_index
        if x_lag is not None:
            self.x_lag = x_lag
        else:
            self.x_lag = 2**int(np.ceil(np.log2( 8*self.m )))
        self._BFS = 0

    def fit(self, X, y=None):
        """Fit detector. y is ignored in unsupervised methods.
        
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The input samples.
        y : Ignored
            Not used, present for API consistency by convention.
        
        Returns
        -------
        self : object
            Fitted estimator.
        """

        self._pv = np.ones(len(X) - self.m + 1, dtype=int)
        aMP = np.zeros_like(self._pv, dtype=float)

        for i in range(self.sp_index, len(X) - self.m + 1):
            if not self._pv[i]:
                aMP[i] = aMP[i-1]
            else:
                aMP[i] = self._BackwardProcessing(X, i)
                self._ForwardProcessing(X, i)

        self.decision_scores_ = aMP
        return self


    def _BackwardProcessing(self, X, i):
        aMP_i = np.inf
        prefix = 2** int(np.ceil(np.log2(self.m)))
        max_lag = min(self.x_lag or i, i)
        reference_ts = X[i-max_lag:i]
        first = True
        expansion_num = 0

        while aMP_i >= self._BFS:
            if prefix >= max_lag:
                aMP_i = min(st.core.mass(X[i:i+self.m], reference_ts))
                if aMP_i > self._BFS:
                    self._BFS = aMP_i
                break
            else:
                if first:
                    first = False
                    aMP_i = min(st.core.mass(X[i:i+self.m], reference_ts[-prefix:]))
                else:
                    start = i-max_lag+(expansion_num * self.m)
                    end = int(i-(max_lag/2)+(expansion_num * self.m))
                    aMP_i = min(st.core.mass(X[i:i+self.m], X[start:end]))

                if aMP_i < self._BFS:
                    break
                else:
                    prefix = 2*prefix
                    expansion_num *= 1

        return aMP_i

    def _ForwardProcessing(self, X, i):
        start = i + self.m
        lookahead = 2** int(np.ceil(np.log2(self.m)))
        end = start + lookahead
        indices = []

        if end < len(X):
            d = st.core.mass(X[i:i+self.m], X[start:end])
            indices = np.argwhere(d < self._BFS)
            indices += start

        self._pv[indices] = 0



import numpy as np
from .utils.metrics import metricor


def generate_curve(label,score,slidingWindow, version='opt', thre=250):
    if version =='opt_mem':
        tpr_3d, fpr_3d, prec_3d, window_3d, avg_auc_3d, avg_ap_3d = metricor().RangeAUC_volume_opt_mem(labels_original=label, score=score, windowSize=slidingWindow, thre=thre)
    else:
        tpr_3d, fpr_3d, prec_3d, window_3d, avg_auc_3d, avg_ap_3d = metricor().RangeAUC_volume_opt(labels_original=label, score=score, windowSize=slidingWindow, thre=thre)
        
    X = np.array(tpr_3d).reshape(1,-1).ravel()
    X_ap = np.array(tpr_3d)[:,:-1].reshape(1,-1).ravel()
    Y = np.array(fpr_3d).reshape(1,-1).ravel()
    W = np.array(prec_3d).reshape(1,-1).ravel()
    Z = np.repeat(window_3d, len(tpr_3d[0]))
    Z_ap = np.repeat(window_3d, len(tpr_3d[0])-1)
    
    return Y, Z, X, X_ap, W, Z_ap,avg_auc_3d, avg_ap_3d


def get_metrics(score, labels, metric='all', version='opt', slidingWindow=5, thre=250):
    """Compute all (or some) evaluation measures for a given score and labels.
    
    Parameters
    ----------
    score : numpy array of shape (n_samples,)
        The input score to evaluate.
    labels : numpy array of shape (n_samples,)
        the labels to compare the score with. It have to be composed of 0 and 1 (1 indicating if the point is an anomaly).
    slidingWindow: int, 
        Buffer length for Range-based AUC measures and for VUS-based measures. 
        For Range-AUC, the buffer length 
        is exactly equals to slidingWindow. For VUS-based measures, the buffer length varies from 0 to ``2*slidingWindow``.
    version : string, optional, default='opt'
        Implementation of VUS.

        - if 'opt', run the default implementation
        - if 'opt_mem', run the optimized implementation, but more complex in memory

    thre : int, optional, default=250
        Number of thresholds for VUS

    metric : string, optional, default='vus'
        compute a subset or all metrics:
        
        - if 'vus', compute and store in a dictionary the following measures (the string is the key of each measure). 

            - 'R_AUC_ROC', Range-adapted version of AUC-ROC [Paparrizos et al. 2022]
            - 'R_AUC_PR', Range-adapted version of AUC-PR [Paparrizos et al. 2022]
            - 'VUS_ROC', Volume under the surface for ROC [Paparrizos et al. 2022]
            - 'VUS_PR', Volume under the surface for PR [Paparrizos et al. 2022]

        - if 'all', compute all measures and store them in a dictionary (the string is the key of each measure). The threhold based measures are computed with a predifined threshold (``score_mu + 3*score_sigma``). In total here are the measures:

            - 'AUC_ROC', Area Under the ROC Curve
            - 'AUC_PR', Area Under the Precision-Recall Curve
            - 'Precision', generic Precision
            - 'Recall', generic Recall
            - 'F', generic F-score (with beta equals to 1)
            - 'Precision_at_k', generic precision at k.
            - 'Rprecision', Time series-adapted Precision [Tatbul et al. 2018].
            - 'Rrecall', Time series-adapted Recall [Tatbul et al. 2018].
            - 'RF', Time series-adapted F-score (with beta equals to 1) [Tatbul et al. 2018].
            - 'R_AUC_ROC', Range-adapted version of AUC-ROC [Paparrizos et al. 2022]
            - 'R_AUC_PR', Range-adapted version of AUC-PR [Paparrizos et al. 2022]
            - 'VUS_ROC', Volume under the surface for ROC [Paparrizos et al. 2022]
            - 'VUS_PR', Volume under the surface for PR [Paparrizos et al. 2022]
            - 'Affiliation_Precision', Affiliation-based precision [Huet et al. 2022]
            - 'Affiliation_Recall', Affiliation-based recall [Huet et al. 2022]
    
    Returns
    -------
    metrics : dictionary
        contains the accuracy values for all evaluation measures.
    """

    metrics = {}
    if metric == 'vus':
        grader = metricor()
        _, _, _, _, _, _,VUS_ROC, VUS_PR = generate_curve(labels, score, slidingWindow, version, thre)

        metrics['VUS_ROC'] = VUS_ROC
        metrics['VUS_PR'] = VUS_PR

        return metrics

    elif metric == 'range_auc':
        grader = metricor()
        R_AUC_ROC, R_AUC_PR, _, _, _ = grader.RangeAUC(labels=labels, score=score, window=slidingWindow, plot_ROC=True)
        
        metrics['R_AUC_ROC'] = R_AUC_ROC
        metrics['R_AUC_PR'] = R_AUC_PR

        return metrics

    elif metric == 'auc':
        
        grader = metricor()
        AUC_ROC = grader.metric_new_auc(labels, score, plot_ROC=False)
        _, _, AUC_PR = grader.metric_PR(labels, score)

        metrics['AUC_ROC'] = AUC_ROC
        metrics['AUC_PR'] = AUC_PR

        return metrics
    
    else :
        from .basic_metrics import basic_metricor
        
        grader = metricor()
        _, _, _, _, _, _,VUS_ROC, VUS_PR = generate_curve(labels, score, slidingWindow, version, thre)
        R_AUC_ROC, R_AUC_PR, _, _, _ = grader.RangeAUC(labels=labels, score=score, window=slidingWindow, plot_ROC=True)
        grader = basic_metricor()
        AUC_ROC, Precision, Recall, F, Rrecall, ExistenceReward, OverlapReward, Rprecision, RF, Precision_at_k = grader.metric_new(labels, score, plot_ROC=False)
        _, _, AUC_PR = grader.metric_PR(labels, score)

        from .affiliation.generics import convert_vector_to_events
        from .affiliation.metrics import pr_from_events

        discrete_score = np.array(score > 0.5, dtype=np.float32)
        events_pred = convert_vector_to_events(discrete_score)
        events_gt = convert_vector_to_events(labels)
        Trange = (0, len(discrete_score))
        affiliation_metrics = pr_from_events(events_pred, events_gt, Trange)
        metrics['AUC_ROC'] = AUC_ROC
        metrics['AUC_PR'] = AUC_PR
        metrics['Precision'] = Precision
        metrics['Recall'] = Recall
        metrics['F'] = F
        metrics['Precision_at_k'] = Precision_at_k
        metrics['Rprecision'] = Rprecision
        metrics['Rrecall'] = Rrecall
        metrics['RF'] = RF
        metrics['R_AUC_ROC'] = R_AUC_ROC
        metrics['R_AUC_PR'] = R_AUC_PR
        metrics['VUS_ROC'] = VUS_ROC
        metrics['VUS_PR'] = VUS_PR
        metrics['Affiliation_Precision'] = affiliation_metrics['Affiliation_Precision']
        metrics['Affiliation_Recall'] = affiliation_metrics['Affiliation_Recall']
        
        if metric == 'all':
            return metrics
        else:
            return metrics[metric]



import numpy as np
from sklearn.preprocessing import MinMaxScaler
from tensorflow.keras.models import Sequential
from tensorflow.keras.callbacks import EarlyStopping
from tensorflow.keras import layers


class AE_MLP2:
    """
    Implementation of AE_MLP2
    
    Parameters
    ----------
    slidingwindow : int
        Subsequence length to analyze.
    epochs : int, (default=10)
        Number of epochs for the training phase

    Attributes
    ----------
    decision_scores_ : numpy array of shape (n_samples - subsequence_length,)
        The anomaly score.
        The higher, the more abnormal. Anomalies tend to have higher
        scores. This value is available once the detector is
        fitted.
    """  
    def __init__(self, slidingWindow = 100, epochs = 10, verbose=0):
        self.slidingWindow = slidingWindow
        self.epochs = epochs
        self.verbose = verbose
        self.model_name = 'AE_MLP2'

    def fit(self, X_clean, X_dirty, ratio = 0.15):
        """Fit detector.
        
        Parameters
        ----------
        X_clean : numpy array of shape (n_samples, )
            The input training samples.
        X_dirty : numpy array of shape (n_samples, )
            The input testing samples.
        ratio : flaot, ([0,1])
            The ratio for the train validation split
        
        Returns
        -------
        self : object
            Fitted estimator.
        """

        TIME_STEPS =  self.slidingWindow
        epochs = self.epochs
        

        X_train = self.create_dataset(X_clean,TIME_STEPS)
        X_test = self.create_dataset(X_dirty,TIME_STEPS)
        
        X_train = MinMaxScaler().fit_transform(X_train.T).T
        X_test = MinMaxScaler().fit_transform(X_test.T).T

        model = Sequential()
        model.add(layers.Dense(32,  activation='relu'))
        model.add(layers.BatchNormalization())
        model.add(layers.Dense(16, activation='relu'))
        model.add(layers.BatchNormalization())
        model.add(layers.Dense(32, activation='relu'))
        model.add(layers.Dense(TIME_STEPS, activation='relu'))
 
        model.compile(optimizer='adam', loss='mse')     
        
        history = model.fit(X_train, X_train,
                        epochs=epochs,
                        batch_size=64,
                        shuffle=False,
                        validation_split=0.15,verbose=self.verbose,
                        callbacks=[EarlyStopping(monitor="val_loss", verbose=self.verbose, patience=5, mode="min")])

        test_predict = model.predict(X_test)
        test_mae_loss = np.mean(np.abs(test_predict - X_test), axis=1)
        nor_test_mae_loss = MinMaxScaler().fit_transform(test_mae_loss.reshape(-1,1)).ravel()
        score = np.zeros(len(X_dirty))
        score[self.slidingWindow//2:self.slidingWindow//2+len(test_mae_loss)]=nor_test_mae_loss
        score[:self.slidingWindow//2]=nor_test_mae_loss[0]
        score[self.slidingWindow//2+len(test_mae_loss):]=nor_test_mae_loss[-1]
        
        self.decision_scores_ = score
        
        return self
        
    
    # Generated training sequences for use in the model.
    def create_dataset(self, X, time_steps):
        output = []
        for i in range(len(X) - time_steps + 1):
            output.append(X[i : (i + time_steps)])
        return np.stack(output)


"""Polynomial Anomaly Detector with GARCH method and raw error method
"""
# Author: Yinchen Wu <yinchen@uchicago.edu>

import numpy as np
import math
from scipy.special import erf
from sklearn.preprocessing import MinMaxScaler
from sklearn.utils.validation import check_is_fitted
from ..models.detectorA import DetectorA


class POLY(DetectorA):
    """Polynomial Anomaly Detector with GARCH method and raw error method
    
    Parameters
    ----------
    Power : int, optional (default=1)
        The power of polynomial fitted to the data
    neighborhood : int, optional (default=max (100, 10*window size))
        The number of samples to fit for one subsequence. Since the timeseries may vary, 
        to caculate the score for the subsequnece (a, a+k) of samples k, we only fit the
        polynomal on its neighborhood.  
    window: int, optional (default = 20)
        The length of the window to detect the given anomolies 
    contamination : float in (0., 0.55), optional (default=0.1)
        The amount of contamination of the data set, i.e. the proportion
        of outliers in the data set. Used when fitting to define the threshold
        on the decision function.

    Attributes
    ----------
    decision_scores_ : numpy array of shape (n_samples,)
        The outlier scores of the training data.
        The higher, the more abnormal. Outliers tend to have higher
        scores. This value is available once the detector is
        fitted.
    estimators_ : dictionary of coefficients at each polynomial
        The collection of fitted sub-estimators.
    threshold_ : float
        The threshold is based on ``contamination``. It is the
        ``n_samples * contamination`` most abnormal samples in
        ``decision_scores_``. The threshold is calculated for generating
        binary outlier labels.
    labels_ : int, either 0 or 1
        The binary labels of the training data. 0 stands for inliers
        and 1 for outliers/anomalies. It is generated by applying
        ``threshold_`` on ``decision_scores_``.
    """
    def __init__(self, power = 1, window = 200, neighborhood = None, contamination = 0.1):
        self.power = power
        self.window = window
        self.avg_window = None
        if neighborhood == None:
            self.neighborhood = max(10*window, 100)
        else:
            self.neighborhood = neighborhood
            
        self.contamination = contamination
        self.model_name = 'POLY'

    def fit(self, X, y=None):
        """Fit detector. y is ignored in unsupervised methods.
        
        Parameters
        ----------
        X : numpy array of shape (n_samples, )
            The input samples.
        y : Ignored
            Not used, present for API consistency by convention.
        
        Returns
        -------
        self : object
            Fitted estimator.
        """
        # validate inputs X and y (optional)
        self._set_n_classes(y)

        self.X_train_ = X
        self.n_train_ = len(X)
        self.decision_scores_ = np.zeros([self.n_train_, 1])
        
        self.n_initial_ = min(500, int(0.1 * self.n_train_))
        self.X_initial_ = X[:self.n_initial_]                         

        window = self.window
        
        if self.avg_window != None:
            self.neighborhood = max(self.neighborhood, 6*self.avg_window) 
            
        if self.neighborhood > len(X):
            self.neighborhood = len(X)
        
        neighborhood = self.neighborhood
                                          
        N = math.floor(self.n_train_ / window)
        M = math.ceil(self.n_initial_ / window)
        data = self.X_train_
        power=self.power
        fit = {}

        for i in range(M, N+1):

            index = int(i * window)    
            neighbor = int(neighborhood/2)

            if index + neighbor < self.n_train_ and index - neighbor > 0: 

                x = np.concatenate((np.arange(index - neighbor, index), np.arange(index + window, index + neighbor)))
                y = np.concatenate((data[index - neighbor: index], data[index + window: index + neighbor] ))
                mymodel = np.poly1d(np.polyfit(x, y, power))
                fit['model' + str(index)] = mymodel
            elif index + neighbor >= self.n_train_ and index + window < self.n_train_:
                x = np.concatenate((np.arange(self.n_train_ - neighborhood, index), np.arange(index + window, self.n_train_)))
                y = np.concatenate((data[self.n_train_ - neighborhood: index], data[index + window: self.n_train_] ))
                mymodel = np.poly1d(np.polyfit(x, y, power))
                fit['model' + str(index)] = mymodel     
            elif index + window >= self.n_train_:
                x = np.arange(self.n_train_ - neighborhood, index)
                y = data[self.n_train_ - neighborhood: index]
                mymodel = np.poly1d(np.polyfit(x, y, power))
                fit['model' + str(index)] = mymodel     
            elif index + window < neighborhood:
                x = np.concatenate((np.arange(0, index), np.arange(index + window, neighborhood)))
                y = np.concatenate((data[0: index], data[index + window: neighborhood] ))
                try:
                    mymodel = np.poly1d(np.polyfit(x, y, power))
                except:
                    x = np.concatenate((np.arange(0, index), np.arange(len(data[index + window: neighborhood]))))
                    mymodel = np.poly1d(np.polyfit(x, y, power))

                fit['model' + str(index)] = mymodel
            else:
                x = np.concatenate((np.arange(index - neighbor, index), np.arange(index + window, index + neighbor)))
                y = np.concatenate((data[index - neighbor: index], data[index + window: index + neighbor] ))
                print(data.shape)
                print(x.shape)
                print(y.shape)
                mymodel = np.poly1d(np.polyfit(x, y, power))
                fit['model' + str(index)] = mymodel
    
        Y = np.zeros(self.n_train_ )
        for i in range(M, N):
            myline = np.linspace(window *i, window * (i+1), window)
            Y[window *i: window * (i+1)] = fit['model' + str(i * window)](myline)
        if self.n_train_ % N != 0:
            x = np.arange(N*window, self.n_train_ )
            Y[N*window:] = fit['model'+str(N * window)](x)  
        self.estimation = Y
        self.estimator = fit
        return self
        
        
    
    def decision_function(self, X= False, measure = None):
        """Derive the decision score based on the given distance measure
        
        Parameters
        ----------
        X : numpy array of shape (n_samples, )
            The input samples.
        measure : object
            object for given distance measure with methods to derive the score
        
        Returns
        -------
        self : object
            Fitted estimator.
        """
        if type(X) != bool:
            self.X_train_ = X
        estimation = self.estimation
        window = self.window
        n_train_ = self.n_train_
        score = np.zeros(n_train_)
        N = math.floor((self.n_train_) / window)
        M = math.ceil(self.n_initial_ / window)
        measure.detector = self
        measure.set_param()

        for i in range(M, N+1):
            index = i * window
            if i == N:
                end = self.n_train_
            else:
                end = index + window
            score[index: index+window] = measure.measure(self.X_train_[index: end], estimation[index: end], index)
        self.decision_scores_ = score
        self._mu = np.mean(self.decision_scores_)
        self._sigma = np.std(self.decision_scores_)
        return self

    def predict_proba(self, X, method='linear', measure = None):
        """Predict the probability of a sample being outlier. Two approaches
        are possible:
        
        1. simply use Min-max conversion to linearly transform the outlier
           scores into the range of [0,1]. The model must be
           fitted first.
        2. use unifying scores.
        
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The input samples.
        method : str, optional, default='linear'
            probability conversion method. It must be one of
            'linear' or 'unify'.
        
        Returns
        -------
        outlier_probability : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. Return the outlier probability, ranging
            in [0,1].
        """

        check_is_fitted(self, ['decision_scores_', 'threshold_', 'labels_'])
        train_scores = self.decision_scores_

        self.fit(X)
        self.decision_function(measure = measure)
        test_scores = self.decision_scores_

        probs = np.zeros([X.shape[0], int(self._classes)])
        if method == 'linear':
            scaler = MinMaxScaler().fit(train_scores.reshape(-1, 1))
            probs[:, 1] = scaler.transform(
                test_scores.reshape(-1, 1)).ravel().clip(0, 1)
            probs[:, 0] = 1 - probs[:, 1]
            return probs

        elif method == 'unify':
            # turn output into probability
            pre_erf_score = (test_scores - self._mu) / (
                    self._sigma * np.sqrt(2))
            erf_score = erf(pre_erf_score)
            probs[:, 1] = erf_score.clip(0, 1).ravel()
            probs[:, 0] = 1 - probs[:, 1]
            return probs
        else:
            raise ValueError(method,
                             'is not a valid probability conversion method')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from ._integral_interval import interval_intersection

def t_start(j, Js = [(1,2),(3,4),(5,6)], Trange = (1,10)):
    """
    Helper for `E_gt_func`
    
    :param j: index from 0 to len(Js) (included) on which to get the start
    :param Js: ground truth events, as a list of couples
    :param Trange: range of the series where Js is included
    :return: generalized start such that the middle of t_start and t_stop 
    always gives the affiliation zone
    """
    b = max(Trange)
    n = len(Js)
    if j == n:
        return(2*b - t_stop(n-1, Js, Trange))
    else:
        return(Js[j][0])

def t_stop(j, Js = [(1,2),(3,4),(5,6)], Trange = (1,10)):
    """
    Helper for `E_gt_func`
    
    :param j: index from 0 to len(Js) (included) on which to get the stop
    :param Js: ground truth events, as a list of couples
    :param Trange: range of the series where Js is included
    :return: generalized stop such that the middle of t_start and t_stop 
    always gives the affiliation zone
    """
    if j == -1:
        a = min(Trange)
        return(2*a - t_start(0, Js, Trange))
    else:
        return(Js[j][1])

def E_gt_func(j, Js, Trange):
    """
    Get the affiliation zone of element j of the ground truth
    
    :param j: index from 0 to len(Js) (excluded) on which to get the zone
    :param Js: ground truth events, as a list of couples
    :param Trange: range of the series where Js is included, can 
    be (-math.inf, math.inf) for distance measures
    :return: affiliation zone of element j of the ground truth represented
    as a couple
    """
    range_left = (t_stop(j-1, Js, Trange) + t_start(j, Js, Trange))/2
    range_right = (t_stop(j, Js, Trange) + t_start(j+1, Js, Trange))/2
    return((range_left, range_right))

def get_all_E_gt_func(Js, Trange):
    """
    Get the affiliation partition from the ground truth point of view
    
    :param Js: ground truth events, as a list of couples
    :param Trange: range of the series where Js is included, can 
    be (-math.inf, math.inf) for distance measures
    :return: affiliation partition of the events
    """
    # E_gt is the limit of affiliation/attraction for each ground truth event
    E_gt = [E_gt_func(j, Js, Trange) for j in range(len(Js))]
    return(E_gt)

def affiliation_partition(Is = [(1,1.5),(2,5),(5,6),(8,9)], E_gt = [(1,2.5),(2.5,4.5),(4.5,10)]):
    """
    Cut the events into the affiliation zones
    The presentation given here is from the ground truth point of view,
    but it is also used in the reversed direction in the main function.
    
    :param Is: events as a list of couples
    :param E_gt: range of the affiliation zones
    :return: a list of list of intervals (each interval represented by either 
    a couple or None for empty interval). The outer list is indexed by each
    affiliation zone of `E_gt`. The inner list is indexed by the events of `Is`.
    """
    out = [None] * len(E_gt)
    for j in range(len(E_gt)):
        E_gt_j = E_gt[j]
        discarded_idx_before = [I[1] < E_gt_j[0] for I in Is]  # end point of predicted I is before the begin of E
        discarded_idx_after = [I[0] > E_gt_j[1] for I in Is] # start of predicted I is after the end of E
        kept_index = [not(a or b) for a, b in zip(discarded_idx_before, discarded_idx_after)]
        Is_j = [x for x, y in zip(Is, kept_index)]
        out[j] = [interval_intersection(I, E_gt[j]) for I in Is_j]
    return(out)



"""Classes of feature mapping for model type B
"""
# Author: Yinchen Wu <yinchen@uchicago.edu>

import numpy as np
import pandas as pd
import math
from builtins import range


class Window:
    """
    Construct a pandas DataFrame of shape (n_samples-m,m) corresponding to all consecutive subsequences in the time series
    
    Parameters
    ----------
    window : int
        Subsequence length.
    """  

    def __init__(self,  window = 100):
        self.window = window
    
    def convert(self, X):
        """Convert the time series X into a pandas DataFrame of shape (n_samples-m,m) corresponding to all consecutive subsequences in the time series.
        
        Parameters
        ----------
        X : numpy array of shape (n_samples, )
            The time series to be transformed.
        
        Returns
        -------
        df : pandas DataFrame
            all consecutive subsequences (of length window) in the time series.
        """

        n = self.window
        X = pd.Series(X)
        L = []
        if n == 0:
            df = X
        else:
            for i in range(n):
                L.append(X.shift(i))
            df = pd.concat(L, axis = 1)
            df = df.iloc[n-1:]
        return df




from __future__ import division
from __future__ import print_function

import warnings
from collections import defaultdict

from ..utils.utility import _get_sklearn_version

#if _get_sklearn_version() > 20:
from inspect import signature
#else:
#    from sklearn.externals.funcsigs import signature

import abc
import six

import numpy as np
from numpy import percentile
from scipy.special import erf
from sklearn.preprocessing import MinMaxScaler
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.multiclass import check_classification_targets
from .sklearn_base import _pprint


@six.add_metaclass(abc.ABCMeta)
class DetectorA(object):
    """Abstract class for type A anomoly detection algorithms.
    Parameters
    ----------
    contamination : float, optional (default=0.1)
        The proportion of anomolies of the data set,
        Used for defining the threshold on the decision function.
    window: int, optional (default = 20)
        The length of the window to detect the given anomolies 
    Attributes
    ----------
    decision_scores_ : numpy array of shape (n_samples,)
        The outlier scores of the training data.
        The higher, the more abnormal. Outliers tend to have higher
        scores. This value is available once the detector is fitted.
    threshold_ : float
        The threshold is based on ``contamination``. It is the
        ``n_samples * contamination`` most abnormal samples in
        ``decision_scores_``. The threshold is calculated for generating
        binary outlier labels.
    labels_ : int, either 0 or 1
        The binary labels of the training data. 0 stands for inliers
        and 1 for outliers/anomalies. It is generated by applying
        ``threshold_`` on ``decision_scores_``.
    """

    @abc.abstractmethod
    def __init__(self, contamination=0.1, window = 75):

        if not (0. < contamination <= 0.55):
            raise ValueError("contamination must be in (0, 0.55], "
                             "got: %f" % contamination)

        self.contamination = contamination

    # noinspection PyIncorrectDocstring
    @abc.abstractmethod
    def fit(self, X, y=None):
        """Fit detector. y is ignored in unsupervised methods.
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The input samples.
        y : Ignored
            Not used, present for API consistency by convention.
        Returns
        -------
        self : object
            Fitted estimator.
        """
        pass

    @abc.abstractmethod
    def decision_function(self, X, measure = None):
        """Predict raw anomaly scores of X using the fitted detector and distance measure
        The anomaly score of an input sample is computed based on the fitted
        detector. For consistency, outliers are assigned with
        higher anomaly scores.
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The input samples. Sparse matrices are accepted only
            if they are supported by the base estimator.
        measre : function class that maps two 1d numpy array to integer. 
            Using euclidean measure if not specified. 
        Returns
        -------
        anomaly_scores : numpy array of shape (n_samples,)
            The anomaly score of the input samples.
        """
        pass

    def predict(self, X):
        """Predict if a particular sample is an outlier or not.
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The input samples.
        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        check_is_fitted(self, ['decision_scores_', 'threshold_', 'labels_'])

        pred_score = self.decision_function(X)
        return (pred_score > self.threshold_).astype('int').ravel()

    def predict_proba(self, method='linear'):
        """Predict the probability of a sample being outlier. Two approaches
        are possible:
        1. simply use Min-max conversion to linearly transform the outlier
           scores into the range of [0,1]. The model must be
           fitted first.
        2. use unifying scores, see :cite:`kriegel2011interpreting`.
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The input samples.
        method : str, optional (default='linear')
            probability conversion method. It must be one of
            'linear' or 'unify'.
        Returns
        -------
        outlier_probability : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. Return the outlier probability, ranging
            in [0,1].
        """

        check_is_fitted(self, ['decision_scores_', 'threshold_', 'labels_'])
        train_scores = self.train_decision_scores_

        test_scores = self.decision_scores_

        probs = np.zeros([self.n_train_, int(self._classes)])
        if method == 'linear':
            scaler = MinMaxScaler().fit(train_scores.reshape(-1, 1))
            probs[:, 1] = scaler.transform(
                test_scores.reshape(-1, 1)).ravel().clip(0, 1)
            probs[:, 0] = 1 - probs[:, 1]
            return probs

        elif method == 'unify':
            # turn output into probability
            pre_erf_score = (test_scores - self._mu) / (
                    self._sigma * np.sqrt(2))
            erf_score = erf(pre_erf_score)
            probs[:, 1] = erf_score.clip(0, 1).ravel()
            probs[:, 0] = 1 - probs[:, 1]
            return probs
        else:
            raise ValueError(method,
                             'is not a valid probability conversion method')

    def _predict_rank(self, normalized=False):
        """Predict the outlyingness rank of a sample by a fitted model. The
        method is for outlier detector score combination.
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The input samples.
        normalized : bool, optional (default=False)
            If set to True, all ranks are normalized to [0,1].
        Returns
        -------
        ranks : array, shape (n_samples,)
            Outlying rank of a sample according to the training data.
        """

        check_is_fitted(self, ['decision_scores_'])

        train_scores = self.train_decision_scores_

        test_scores = self.decision_scores_

        sorted_train_scores = np.sort(train_scores)
        ranks = np.searchsorted(sorted_train_scores, test_scores)

        if normalized:
            # return normalized ranks
            ranks = ranks / ranks.max()
        return ranks

 

    def _set_n_classes(self, y):
        """Set the number of classes if `y` is presented, which is not
        expected. It could be useful for multi-class outlier detection.
        Parameters
        ----------
        y : numpy array of shape (n_samples,)
            Ground truth.
        Returns
        -------
        self
        """

        self._classes = 2  # default as binary classification
        if y is not None:
            check_classification_targets(y)
            self._classes = len(np.unique(y))
            warnings.warn(
                "y should not be presented in unsupervised learning.")
        return self

    
    def _process_decision_scores(self):
        """Internal function to calculate key attributes:
        - threshold_: used to decide the binary label
        - labels_: binary labels of training data
        Returns
        -------
        self
        """

        self.threshold_ = percentile(self.decision_scores_,
                                     100 * (1 - self.contamination))
        self.labels_ = (self.decision_scores_ > self.threshold_).astype(
            'int').ravel()

        # calculate for predict_proba()

        self._mu = np.mean(self.decision_scores_)
        self._sigma = np.std(self.decision_scores_)

        return self

    def _set_threshold(self, X_train, measure = None):
        """Internal function to calculate key attributes:
        - threshold_: used to decide the binary label
        - labels_: binary labels of training data
        - measure: the measure class to define the set
        Returns
        -------
        self
        """
        self.fit(X_train)
        self.decision_function(measure = measure)

        self.train_decision_scores_ = self.decision_scores_
    
        self.threshold_ = percentile(self.decision_scores_,
                                     100 * (1 - self.contamination))
        self.labels_ = (self.decision_scores_ > self.threshold_).astype(
            'int').ravel()

        # calculate for predict_proba()

        self._mu = np.mean(self.decision_scores_)
        self._sigma = np.std(self.decision_scores_)

        return self
    # noinspection PyMethodParameters
    def _get_param_names(cls):
        # noinspection PyPep8
        """Get parameter names for the estimator
        See http://scikit-learn.org/stable/modules/generated/sklearn.base.BaseEstimator.html
        and sklearn/base.py for more information.
        """

        # fetch the constructor or the original constructor before
        # deprecation wrapping if any
        init = getattr(cls.__init__, 'deprecated_original', cls.__init__)
        if init is object.__init__:
            # No explicit constructor to introspect
            return []

        # introspect the constructor arguments to find the model parameters
        # to represent
        init_signature = signature(init)
        # Consider the constructor parameters excluding 'self'
        parameters = [p for p in init_signature.parameters.values()
                      if p.name != 'self' and p.kind != p.VAR_KEYWORD]
        for p in parameters:
            if p.kind == p.VAR_POSITIONAL:
                raise RuntimeError("scikit-learn estimators should always "
                                   "specify their parameters in the signature"
                                   " of their __init__ (no varargs)."
                                   " %s with constructor %s doesn't "
                                   " follow this convention."
                                   % (cls, init_signature))
        # Extract and sort argument names excluding 'self'
        return sorted([p.name for p in parameters])

    # noinspection PyPep8
    def get_params(self, deep=True):
        """Get parameters for this estimator.
        See http://scikit-learn.org/stable/modules/generated/sklearn.base.BaseEstimator.html
        and sklearn/base.py for more information.
        Parameters
        ----------
        deep : bool, optional (default=True)
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.
        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        """

        out = dict()
        for key in self._get_param_names():
            # We need deprecation warnings to always be on in order to
            # catch deprecated param values.
            # This is set in utils/__init__.py but it gets overwritten
            # when running under python3 somehow.
            warnings.simplefilter("always", DeprecationWarning)
            try:
                with warnings.catch_warnings(record=True) as w:
                    value = getattr(self, key, None)
                if len(w) and w[0].category == DeprecationWarning:
                    # if the parameter is deprecated, don't show it
                    continue
            finally:
                warnings.filters.pop(0)

            # XXX: should we rather test if instance of estimator?
            if deep and hasattr(value, 'get_params'):
                deep_items = value.get_params().items()
                out.update((key + '__' + k, val) for k, val in deep_items)
            out[key] = value
        return out

    def set_params(self, **params):
        # noinspection PyPep8
        """Set the parameters of this estimator.
        The method works on simple estimators as well as on nested objects
        (such as pipelines). The latter have parameters of the form
        ``<component>__<parameter>`` so that it's possible to update each
        component of a nested object.
        See http://scikit-learn.org/stable/modules/generated/sklearn.base.BaseEstimator.html
        and sklearn/base.py for more information.
        Returns
        -------
        self : object
        """

        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        valid_params = self.get_params(deep=True)

        nested_params = defaultdict(dict)  # grouped by prefix
        for key, value in params.items():
            key, delim, sub_key = key.partition('__')
            if key not in valid_params:
                raise ValueError('Invalid parameter %s for estimator %s. '
                                 'Check the list of available parameters '
                                 'with `estimator.get_params().keys()`.' %
                                 (key, self))

            if delim:
                nested_params[key][sub_key] = value
            else:
                setattr(self, key, value)

        for key, sub_params in nested_params.items():
            valid_params[key].set_params(**sub_params)

        return self

    def __repr__(self):
        # noinspection PyPep8
        """
        See http://scikit-learn.org/stable/modules/generated/sklearn.base.BaseEstimator.html
        and sklearn/base.py for more information.
        """

        class_name = self.__class__.__name__
        return '%s(%s)' % (class_name, _pprint(self.get_params(deep=False),
                                               offset=len(class_name), ),)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from itertools import groupby
from operator import itemgetter
import math
import gzip
import glob
import os

def convert_vector_to_events(vector = [0, 1, 1, 0, 0, 1, 0]):
    """
    Convert a binary vector (indicating 1 for the anomalous instances)
    to a list of events. The events are considered as durations,
    i.e. setting 1 at index i corresponds to an anomalous interval [i, i+1).
    
    :param vector: a list of elements belonging to {0, 1}
    :return: a list of couples, each couple representing the start and stop of
    each event
    """
    positive_indexes = [idx for idx, val in enumerate(vector) if val > 0]
    events = []
    for k, g in groupby(enumerate(positive_indexes), lambda ix : ix[0] - ix[1]):
        cur_cut = list(map(itemgetter(1), g))
        events.append((cur_cut[0], cur_cut[-1]))
    
    # Consistent conversion in case of range anomalies (for indexes):
    # A positive index i is considered as the interval [i, i+1),
    # so the last index should be moved by 1
    events = [(x, y+1) for (x,y) in events]
        
    return(events)

def infer_Trange(events_pred, events_gt):
    """
    Given the list of events events_pred and events_gt, get the
    smallest possible Trange corresponding to the start and stop indexes 
    of the whole series.
    Trange will not influence the measure of distances, but will impact the
    measures of probabilities.
    
    :param events_pred: a list of couples corresponding to predicted events
    :param events_gt: a list of couples corresponding to ground truth events
    :return: a couple corresponding to the smallest range containing the events
    """
    if len(events_gt) == 0:
        raise ValueError('The gt events should contain at least one event')
    if len(events_pred) == 0:
        # empty prediction, base Trange only on events_gt (which is non empty)
        return(infer_Trange(events_gt, events_gt))
        
    min_pred = min([x[0] for x in events_pred])
    min_gt = min([x[0] for x in events_gt])
    max_pred = max([x[1] for x in events_pred])
    max_gt = max([x[1] for x in events_gt])
    Trange = (min(min_pred, min_gt), max(max_pred, max_gt))
    return(Trange)

def has_point_anomalies(events):
    """
    Checking whether events contain point anomalies, i.e.
    events starting and stopping at the same time.
    
    :param events: a list of couples corresponding to predicted events
    :return: True is the events have any point anomalies, False otherwise
    """
    if len(events) == 0:
        return(False)
    return(min([x[1] - x[0] for x in events]) == 0)

def _sum_wo_nan(vec):
    """
    Sum of elements, ignoring math.isnan ones
    
    :param vec: vector of floating numbers
    :return: sum of the elements, ignoring math.isnan ones
    """
    vec_wo_nan = [e for e in vec if not math.isnan(e)]
    return(sum(vec_wo_nan))
    
def _len_wo_nan(vec):
    """
    Count of elements, ignoring math.isnan ones
    
    :param vec: vector of floating numbers
    :return: count of the elements, ignoring math.isnan ones
    """
    vec_wo_nan = [e for e in vec if not math.isnan(e)]
    return(len(vec_wo_nan))

def read_gz_data(filename = 'data/machinetemp_groundtruth.gz'):
    """
    Load a file compressed with gz, such that each line of the
    file is either 0 (representing a normal instance) or 1 (representing)
    an anomalous instance.
    :param filename: file path to the gz compressed file
    :return: list of integers with either 0 or 1
    """
    with gzip.open(filename, 'rb') as f:
        content = f.read().splitlines()
    content = [int(x) for x in content]
    return(content)

def read_all_as_events():
    """
    Load the files contained in the folder `data/` and convert
    to events. The length of the series is kept.
    The convention for the file name is: `dataset_algorithm.gz`
    :return: two dictionaries:
        - the first containing the list of events for each dataset and algorithm,
        - the second containing the range of the series for each dataset
    """
    filepaths = glob.glob('data/*.gz')
    datasets = dict()
    Tranges = dict()
    for filepath in filepaths:
        vector = read_gz_data(filepath)
        events = convert_vector_to_events(vector)
        # ad hoc cut for those files
        cut_filepath = (os.path.split(filepath)[1]).split('_')
        data_name = cut_filepath[0]
        algo_name = (cut_filepath[1]).split('.')[0]
        if not data_name in datasets:
            datasets[data_name] = dict()
            Tranges[data_name] = (0, len(vector))
        datasets[data_name][algo_name] = events
    return(datasets, Tranges)

def f1_func(p, r):
    """
    Compute the f1 function
    :param p: precision numeric value
    :param r: recall numeric value
    :return: f1 numeric value
    """
    return(2*p*r/(p+r))



from statsmodels.tsa.stattools import acf
from scipy.signal import argrelextrema
import numpy as np



def find_length(data):
    """
    determine sliding window (period) based on autocorrelation.
        
    Parameters
    ----------
    data : numpy array of shape (n_samples, )
        The time series on which we find the optimal subsequence length.
    
    Returns
    -------
    length : int
        argmax on the autocorrelation curve. Cannot be smaller than 3 and bigger than 300.
        In case of extreme small (below 3) or big (above 300) argmax, we set a default subseuqence length and return 100. 
    """
    if len(data.shape)>1:
        return 0
    data = data[:min(20000, len(data))]
    
    base = 3
    auto_corr = acf(data, nlags=400, fft=True)[base:]
    
    
    local_max = argrelextrema(auto_corr, np.greater)[0]
    try:
        max_local_max = np.argmax([auto_corr[lcm] for lcm in local_max])
        if local_max[max_local_max]<3 or local_max[max_local_max]>300:
            return 100
        return local_max[max_local_max]+base
    except:
        return 100
    



from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import LSTM
from tensorflow.keras.callbacks import EarlyStopping


import numpy as np
import math
from sklearn.utils.validation import check_is_fitted
from sklearn.preprocessing import MinMaxScaler

class lstm:
    """
    Implementation of LSTM-AD
    
    Parameters
    ----------
    slidingwindow : int
        Subsequence length to analyze.
    predict_time_steps : int, (default=1)
        The length of the subsequence to predict.
    epochs : int, (default=10)
        Number of epochs for the training phase
    patience : int, (default=10)
        Number of epoch to wait before early stopping during training

    Attributes
    ----------
    decision_scores_ : numpy array of shape (n_samples - subsequence_length,)
        The anomaly score.
        The higher, the more abnormal. Anomalies tend to have higher
        scores. This value is available once decision_function is called.
    """

    def __init__(self, slidingwindow = 100, predict_time_steps=1, epochs = 10, patience = 10, verbose=0):
        self.slidingwindow = slidingwindow
        self.predict_time_steps = predict_time_steps
        self.epochs = epochs
        self.patience = patience
        self.verbose = verbose
        self.model_name = 'LSTM2'
        
    def fit(self, X_clean, X_dirty, ratio = 0.15):
        """Fit detector.
        
        Parameters
        ----------
        X_clean : numpy array of shape (n_samples, )
            The input training samples.
        X_dirty : numpy array of shape (n_samples, )
            The input testing samples.
        ratio : flaot, ([0,1])
            The ratio for the train validation split
        
        Returns
        -------
        self : object
            Fitted estimator.
        """

        slidingwindow = self.slidingwindow
        predict_time_steps = self.predict_time_steps
        self.n_test_ = len(X_dirty)

        X_train, Y_train = self.create_dataset(X_clean, slidingwindow, predict_time_steps)
        X_test, Y_test = self.create_dataset(X_dirty, slidingwindow, predict_time_steps)
        
        X_train = X_train.reshape((X_train.shape[0], X_train.shape[1], 1))
        X_test = X_test.reshape((X_test.shape[0], X_test.shape[1], 1))
        
        model = Sequential()
        model.add(LSTM(50,return_sequences=True,input_shape=(X_train.shape[1],X_train.shape[2])))
        model.add(LSTM(50))
        model.add(Dense(predict_time_steps))
        model.compile(loss='mean_squared_error', optimizer='adam')
        
        es = EarlyStopping(monitor='val_loss', mode='min', verbose=self.verbose, patience=self.patience)
        
        model.fit(X_train,Y_train,validation_split=ratio,
                  epochs=self.epochs,batch_size=64,verbose=self.verbose, callbacks=[es])
        
        prediction = model.predict(X_test)

        self.Y = Y_test
        self.estimation = prediction
        self.estimator = model
        self.n_initial = X_train.shape[0]
        
        return self
        
    def create_dataset(self, X, slidingwindow, predict_time_steps=1): 
        Xs, ys = [], []
        for i in range(len(X) - slidingwindow - predict_time_steps+1):
            tmp = X[i : i + slidingwindow + predict_time_steps]
            tmp= MinMaxScaler(feature_range=(0,1)).fit_transform(tmp.reshape(-1,1)).ravel()
            x = tmp[:slidingwindow]
            y = tmp[slidingwindow:]
            
            Xs.append(x)
            ys.append(y)
        return np.array(Xs), np.array(ys)
    

    
    def decision_function(self, measure = None):
        """Derive the decision score based on the given distance measure
        
        Parameters
        ----------
        measure : object
            object for given distance measure with methods to derive the score
        
        Returns
        -------
        self : object
            Fitted estimator.
        """
       
        Y_test = self.Y

        score = np.zeros(self.n_test_)
        estimation = self.estimation

        for i in range(estimation.shape[0]):
            score[i - estimation.shape[0]] = measure.measure(Y_test[i], estimation[i], self.n_test_ - estimation.shape[0] + i)

        score[0: - estimation.shape[0]] = score[- estimation.shape[0]]
        
        self.decision_scores_ = score
        return self




from sklearn import metrics
import numpy as np
import math

class metricor:
    def __init__(self, a = 1, probability = True, bias = 'flat', ):
        self.a = a
        self.probability = probability
        self.bias = bias

    def detect_model(self, model, label, contamination = 0.1, window = 100, is_A = False, is_threshold = True):
        if is_threshold:
            score = self.scale_threshold(model.decision_scores_, model._mu, model._sigma)
        else:
            score = self.scale_contamination(model.decision_scores_, contamination = contamination)
        if is_A is False:
            scoreX = np.zeros(len(score)+window)
            scoreX[math.ceil(window/2): len(score)+window - math.floor(window/2)] = score 
        else:
            scoreX = score
        
        self.score_=scoreX
        L = self.metric(label, scoreX)
        return L

        
    def labels_conv(self, preds):
        '''return indices of predicted anomaly
        '''

        # p = np.zeros(len(preds))
        index = np.where(preds >= 0.5)
        return index[0]
    
    def labels_conv_binary(self, preds):
        '''return predicted label
        '''
        p = np.zeros(len(preds))
        index = np.where(preds >= 0.5)
        p[index[0]] = 1
        return p 

    def w(self, AnomalyRange, p):
        MyValue = 0
        MaxValue = 0
        start = AnomalyRange[0]
        AnomalyLength = AnomalyRange[1] - AnomalyRange[0] + 1
        for i in range(start, start +AnomalyLength):
            bi = self.b(i, AnomalyLength)
            MaxValue +=  bi
            if i in p:
                MyValue += bi
        return MyValue/MaxValue

    def Cardinality_factor(self, Anomolyrange, Prange):  ## Changed this is better reduced if conditions
        score = 0 
        start = Anomolyrange[0]
        end = Anomolyrange[1]
        print(start, end)
        for i in Prange:
            if not(i[1]<start or i[0]>end):
                score+=1
        if score == 0:
            return 0
        else:
            return 1/score

    def b(self, i, length):
        bias = self.bias 
        if bias == 'flat':
            return 1
        elif bias == 'front-end bias':
            return length - i + 1
        elif bias == 'back-end bias':
            return i
        else:
            if i <= length/2:
                return i
            else:
                return length - i + 1
            
    def scale_threshold(self, score, score_mu, score_sigma):
        return (score >= (score_mu + 3*score_sigma)).astype(int)

    def metric_new(self, label, score, plot_ROC=False, alpha=0.2,coeff=3):
        if np.sum(label) == 0:
            print('All labels are 0.')
            return None

        if np.isnan(score).any() or score is None:
            print('Score must not be none.')
            return None

        #auc-roc1
        auc = metrics.roc_auc_score(label, score)
        if plot_ROC:
            fpr, tpr, thresholds  = metrics.roc_curve(label, score)
            display = metrics.RocCurveDisplay(fpr=fpr, tpr=tpr, roc_auc=auc)
            display.plot()            

        #precision, recall, F
        preds = score > (np.mean(score)+coeff*np.std(score))  
        if not(np.any(preds)):
            preds= score > (np.mean(score)+2*coeff*np.std(score)/3) 
            if np.any(preds):
                preds = score > (np.mean(score)+1*coeff*np.std(score)/3) 
                if np.any(preds):
                    preds = score > (np.mean(score) + 1*np.std(score))
    
        Precision, Recall, F, Support = metrics.precision_recall_fscore_support(label, preds, zero_division=0)
        precision = Precision[1]
        recall = Recall[1]
        f = F[1]

        #range based anomaly 
        Rrecall, ExistenceReward, OverlapReward = self.range_recall_new(label, preds, alpha)
        Rprecision = self.range_recall_new(preds, label, 0)[0]
        
        if Rprecision + Rrecall==0:
            Rf=0
        else:
            Rf = 2 * Rrecall * Rprecision / (Rprecision + Rrecall)
        
        # top-k
        k = int(np.sum(label))
        threshold = np.percentile(score, 100 * (1-k/len(label)))
        
        # precision_at_k = metrics.top_k_accuracy_score(label, score, k)
        p_at_k = np.where(preds > threshold)[0]
        TP_at_k = sum(label[p_at_k])
        precision_at_k = TP_at_k/k
        
        L = np.array([auc, precision, recall, f, Rrecall, ExistenceReward, OverlapReward, Rprecision, Rf, precision_at_k])
        if plot_ROC:
            return L, fpr, tpr
        return L

    def metric_new_auc(self, label, score, plot_ROC=False, alpha=0.2,coeff=3):
        if np.sum(label) == 0:
            print('All labels are 0. Label must have groud truth value for calculating AUC score.')
            return None

        if np.isnan(score).any() or score is None:
            print('Score must not be none.')
            return None

        #area under curve
        auc = metrics.roc_auc_score(label, score)
        
        return auc

    def metric_PR(self, label, score):
        precision, recall, thresholds = metrics.precision_recall_curve(label, score)
        # plt.figure()
        # disp = metrics.PrecisionRecallDisplay(precision=precision, recall=recall)
        # disp.plot()
        AP = metrics.auc(recall, precision)
        #AP = metrics.average_precision_score(label, score)
        return precision, recall, AP
        
    def range_recall_new(self, labels, preds, alpha):   


        p = np.where(preds == 1)[0]    # positions of predicted label==1
        range_pred = self.range_convers_new(preds)  
        range_label = self.range_convers_new(labels)
        
        Nr = len(range_label)    # total # of real anomaly segments

        ExistenceReward = self.existence_reward(range_label, p)


        OverlapReward = 0
        for i in range_label:
            OverlapReward += self.w(i, p) * self.Cardinality_factor(i, range_pred)


        score = alpha * ExistenceReward + (1-alpha) * OverlapReward
        if Nr != 0:
            return score/Nr, ExistenceReward/Nr, OverlapReward/Nr
        else:
            return 0,0,0


    def range_convers_new(self, label):
        '''
        input: arrays of binary values 
        output: list of ordered pair [[a0,b0], [a1,b1]... ] of the inputs
        '''
        L = []
        i = 0
        j = 0 
        while j < len(label):
            # print(i)
            while label[i] == 0:
                i+=1
                if i >= len(label):  #?
                    break            #?
            j = i+1
            # print('j'+str(j))
            if j >= len(label):
                if j==len(label):
                    L.append((i,j-1))
    
                break
            while label[j] != 0:
                j+=1
                if j >= len(label):
                    L.append((i,j-1))
                    break
            if j >= len(label):
                break
            L.append((i, j-1))
            i = j
        return L
    
    def existence_reward(self, labels, preds):
        '''
        labels: list of ordered pair 
        preds predicted data
        '''

        score = 0
        for i in labels:
            if np.sum(np.multiply(preds <= i[1], preds >= i[0])) > 0:
                score += 1
        return score
    
    def num_nonzero_segments(self, x):
        count=0
        if x[0]>0:
            count+=1
        for i in range(1, len(x)):
            if x[i]>0 and x[i-1]==0:
                count+=1
        return count

    def extend_postive_range(self, x, window=5):
        label = x.copy().astype(float)
        L = self.range_convers_new(label)   # index of non-zero segments
        length = len(label)
        for k in range(len(L)):
            s = L[k][0]
            e = L[k][1] 
            
            
            x1 = np.arange(e+1,min(e+window//2+1,length))
            label[x1] += np.sqrt(1 - (x1-e)/(window))
            
            x2 = np.arange(max(s-window//2,0),s)
            label[x2] += np.sqrt(1 - (s-x2)/(window))
            
        label = np.minimum(np.ones(length), label)
        return label

    
    def extend_postive_range_individual(self, x, percentage=0.2):
        label = x.copy().astype(float)
        L = self.range_convers_new(label)   # index of non-zero segments
        length = len(label)
        for k in range(len(L)):
            s = L[k][0] 
            e = L[k][1] 
            
            l0 = int((e-s+1)*percentage)
            
            x1 = np.arange(e,min(e+l0,length))
            label[x1] += np.sqrt(1 - (x1-e)/(2*l0))
            
            x2 = np.arange(max(s-l0,0),s)
            label[x2] += np.sqrt(1 - (s-x2)/(2*l0))
            
        label = np.minimum(np.ones(length), label)
        return label
    
    def TPR_FPR_RangeAUC(self, labels, pred, P, L):
        product = labels * pred
        
        TP = np.sum(product)
        
        # recall = min(TP/P,1)
        P_new = (P+np.sum(labels))/2      # so TPR is neither large nor small
        # P_new = np.sum(labels)
        recall = min(TP/P_new,1)
        # recall = TP/np.sum(labels)
        # print('recall '+str(recall))
        
        
        existence = 0
        for seg in L:
            if np.sum(product[seg[0]:(seg[1]+1)])>0:
                existence += 1
                
        existence_ratio = existence/len(L)
        # print(existence_ratio)
        
        # TPR_RangeAUC = np.sqrt(recall*existence_ratio)
        # print(existence_ratio)
        TPR_RangeAUC = recall*existence_ratio
        
        FP = np.sum(pred) - TP
        # TN = np.sum((1-pred) * (1-labels))
        
        # FPR_RangeAUC = FP/(FP+TN)
        N_new = len(labels) - P_new
        FPR_RangeAUC = FP/N_new
        
        Precision_RangeAUC = TP/np.sum(pred)
        
        return TPR_RangeAUC, FPR_RangeAUC, Precision_RangeAUC
    
    def RangeAUC(self, labels, score, window=0, percentage=0, plot_ROC=False, AUC_type='window'):
        # AUC_type='window'/'percentage'
        score_sorted = -np.sort(-score)
        
        P = np.sum(labels)
        # print(np.sum(labels))
        if AUC_type=='window':
            labels = self.extend_postive_range(labels, window=window)
        else:   
            labels = self.extend_postive_range_individual(labels, percentage=percentage)
        
        # print(np.sum(labels))
        L = self.range_convers_new(labels)
        TF_list = np.zeros((252,2))
        Precision_list = np.ones(251)
        j=0
        for i in np.linspace(0, len(score)-1, 250).astype(int):
            threshold = score_sorted[i]
            # print('thre='+str(threshold))
            pred = score>= threshold
            TPR, FPR, Precision = self.TPR_FPR_RangeAUC(labels, pred, P,L)
            j+=1
            TF_list[j]=[TPR,FPR]
            Precision_list[j]=(Precision)
            
        TF_list[j+1]=[1,1]
                
        width = TF_list[1:,1] - TF_list[:-1,1]
        height = (TF_list[1:,0] + TF_list[:-1,0])/2
        AUC_range = np.dot(width,height)

        width_PR = TF_list[1:-1,0] - TF_list[:-2,0]
        height_PR = (Precision_list[1:] + Precision_list[:-1])/2
        AP_range = np.dot(width_PR,height_PR)
        
        if plot_ROC:
            return AUC_range, AP_range, TF_list[:,1], TF_list[:,0], Precision_list
        
        return AUC_range

    def new_sequence(self, label, sequence_original, window):
        a = max(sequence_original[0][0] - window // 2, 0)
        sequence_new = []
        for i in range(len(sequence_original) - 1):
            if sequence_original[i][1] + window // 2 < sequence_original[i + 1][0] - window // 2:
                sequence_new.append((a, sequence_original[i][1] + window // 2))
                a = sequence_original[i + 1][0] - window // 2
        sequence_new.append((a, min(sequence_original[len(sequence_original) - 1][1] + window // 2, len(label) - 1)))
        return sequence_new

    def sequencing(self, x, L, window=5):
        label = x.copy().astype(float)
        length = len(label)

        for k in range(len(L)):
            s = L[k][0]
            e = L[k][1]

            x1 = np.arange(e + 1, min(e + window // 2 + 1, length))
            label[x1] += np.sqrt(1 - (x1 - e) / (window))

            x2 = np.arange(max(s - window // 2, 0), s)
            label[x2] += np.sqrt(1 - (s - x2) / (window))

        label = np.minimum(np.ones(length), label)
        return label

    # TPR_FPR_window
    def RangeAUC_volume_opt(self, labels_original, score, windowSize, thre=250):
        window_3d = np.arange(0, windowSize + 1, 1)
        P = np.sum(labels_original)
        seq = self.range_convers_new(labels_original)
        l = self.new_sequence(labels_original, seq, windowSize)

        score_sorted = -np.sort(-score)

        tpr_3d = np.zeros((windowSize + 1, thre + 2))
        fpr_3d = np.zeros((windowSize + 1, thre + 2))
        prec_3d = np.zeros((windowSize + 1, thre + 1))

        auc_3d = np.zeros(windowSize + 1)
        ap_3d = np.zeros(windowSize + 1)

        tp = np.zeros(thre)
        N_pred = np.zeros(thre)

        for k, i in enumerate(np.linspace(0, len(score) - 1, thre).astype(int)):
            threshold = score_sorted[i]
            pred = score >= threshold
            N_pred[k] = np.sum(pred)

        for window in window_3d:

            labels_extended = self.sequencing(labels_original, seq, window)
            L = self.new_sequence(labels_extended, seq, window)

            TF_list = np.zeros((thre + 2, 2))
            Precision_list = np.ones(thre + 1)
            j = 0

            for i in np.linspace(0, len(score) - 1, thre).astype(int):
                threshold = score_sorted[i]
                pred = score >= threshold
                labels = labels_extended.copy()
                existence = 0

                for seg in L:
                    labels[seg[0]:seg[1] + 1] = labels_extended[seg[0]:seg[1] + 1] * pred[seg[0]:seg[1] + 1]
                    if (pred[seg[0]:(seg[1] + 1)] > 0).any():
                        existence += 1
                for seg in seq:
                    labels[seg[0]:seg[1] + 1] = 1

                TP = 0
                N_labels = 0
                for seg in l:
                    TP += np.dot(labels[seg[0]:seg[1] + 1], pred[seg[0]:seg[1] + 1])
                    N_labels += np.sum(labels[seg[0]:seg[1] + 1])

                TP += tp[j]
                FP = N_pred[j] - TP

                existence_ratio = existence / len(L)

                P_new = (P + N_labels) / 2
                recall = min(TP / P_new, 1)

                TPR = recall * existence_ratio
                N_new = len(labels) - P_new
                FPR = FP / N_new

                Precision = TP / N_pred[j]

                j += 1
                TF_list[j] = [TPR, FPR]
                Precision_list[j] = Precision

            TF_list[j + 1] = [1, 1]  # otherwise, range-AUC will stop earlier than (1,1)

            tpr_3d[window] = TF_list[:, 0]
            fpr_3d[window] = TF_list[:, 1]
            prec_3d[window] = Precision_list

            width = TF_list[1:, 1] - TF_list[:-1, 1]
            height = (TF_list[1:, 0] + TF_list[:-1, 0]) / 2
            AUC_range = np.dot(width, height)
            auc_3d[window] = (AUC_range)

            width_PR = TF_list[1:-1, 0] - TF_list[:-2, 0]
            height_PR = Precision_list[1:]

            AP_range = np.dot(width_PR, height_PR)
            ap_3d[window] = AP_range

        return tpr_3d, fpr_3d, prec_3d, window_3d, sum(auc_3d) / len(window_3d), sum(ap_3d) / len(window_3d)

    def RangeAUC_volume_opt_mem(self, labels_original, score, windowSize, thre=250):
        window_3d = np.arange(0, windowSize + 1, 1)
        P = np.sum(labels_original)
        seq = self.range_convers_new(labels_original)
        l = self.new_sequence(labels_original, seq, windowSize)

        score_sorted = -np.sort(-score)

        tpr_3d = np.zeros((windowSize + 1, thre + 2))
        fpr_3d = np.zeros((windowSize + 1, thre + 2))
        prec_3d = np.zeros((windowSize + 1, thre + 1))

        auc_3d = np.zeros(windowSize + 1)
        ap_3d = np.zeros(windowSize + 1)

        tp = np.zeros(thre)
        N_pred = np.zeros(thre)
        p = np.zeros((thre, len(score)))

        for k, i in enumerate(np.linspace(0, len(score) - 1, thre).astype(int)):
            threshold = score_sorted[i]
            pred = score >= threshold
            p[k] = pred
            N_pred[k] = np.sum(pred)

        for window in window_3d:
            labels_extended = self.sequencing(labels_original, seq, window)
            L = self.new_sequence(labels_extended, seq, window)

            TF_list = np.zeros((thre + 2, 2))
            Precision_list = np.ones(thre + 1)
            j = 0

            for i in np.linspace(0, len(score) - 1, thre).astype(int):
                labels = labels_extended.copy()
                existence = 0

                for seg in L:
                    labels[seg[0]:seg[1] + 1] = labels_extended[seg[0]:seg[1] + 1] * p[j][seg[0]:seg[1] + 1]
                    if (p[j][seg[0]:(seg[1] + 1)] > 0).any():
                        existence += 1
                for seg in seq:
                    labels[seg[0]:seg[1] + 1] = 1

                N_labels = 0
                TP = 0
                for seg in l:
                    TP += np.dot(labels[seg[0]:seg[1] + 1], p[j][seg[0]:seg[1] + 1])
                    N_labels += np.sum(labels[seg[0]:seg[1] + 1])

                TP += tp[j]
                FP = N_pred[j] - TP

                existence_ratio = existence / len(L)

                P_new = (P + N_labels) / 2
                recall = min(TP / P_new, 1)

                TPR = recall * existence_ratio

                N_new = len(labels) - P_new
                FPR = FP / N_new
                Precision = TP / N_pred[j]
                j += 1

                TF_list[j] = [TPR, FPR]
                Precision_list[j] = Precision

            TF_list[j + 1] = [1, 1]
            tpr_3d[window] = TF_list[:, 0]
            fpr_3d[window] = TF_list[:, 1]
            prec_3d[window] = Precision_list

            width = TF_list[1:, 1] - TF_list[:-1, 1]
            height = (TF_list[1:, 0] + TF_list[:-1, 0]) / 2
            AUC_range = np.dot(width, height)
            auc_3d[window] = (AUC_range)

            width_PR = TF_list[1:-1, 0] - TF_list[:-2, 0]
            height_PR = Precision_list[1:]
            AP_range = np.dot(width_PR, height_PR)
            ap_3d[window] = (AP_range)
        return tpr_3d, fpr_3d, prec_3d, window_3d, sum(auc_3d) / len(window_3d), sum(ap_3d) / len(window_3d)    

    def RangeAUC_volume_opt_time(self, labels_original, score, windowSize, thre=250):
        window_3d = np.arange(0, windowSize+1, 1)
        P = np.sum(labels_original)
        seq = self.range_convers_new(labels_original)
        l = self.new_sequence(labels_original, seq, windowSize)

        score_sorted = -np.sort(-score)

        tpr_3d=np.zeros((windowSize+1,thre+2))
        fpr_3d=np.zeros((windowSize+1,thre+2))
        prec_3d=np.zeros((windowSize+1,thre+1))

        auc_3d=np.zeros(windowSize+1)
        ap_3d=np.zeros(windowSize+1)

        tp = np.zeros(thre)
        N_pred = np.zeros(thre)
        p = np.zeros((thre,len(score)))
        tmp = np.linspace(0, len(score)-1, thre).astype(int)

        for k, i in enumerate(tmp):
            threshold = score_sorted[i]
            pred = score>= threshold
            p[k]=pred
            N_pred[k]=np.sum(pred)

        for window in window_3d:

            labels = self.sequencing(labels_original, seq, window)
            L = self.new_sequence(labels, seq, window)

            TF_list = np.zeros((thre+2,2))
            Precision_list = np.ones(thre+1)
            j=0
            N_labels = 0

            for seg in l:
                N_labels += np.sum(labels[seg[0]:seg[1]+1])

            TF_list, Precision_list, j = vectorized(tmp, score_sorted, score, l, labels, j, tp, N_pred, L, P, N_labels, TF_list, Precision_list)

            TF_list[j+1]=[1,1]
            

            tpr_3d[window]=TF_list[:,0]
            fpr_3d[window]=TF_list[:,1]
            prec_3d[window]=Precision_list
            
            width = TF_list[1:,1] - TF_list[:-1,1]
            height = (TF_list[1:,0] + TF_list[:-1,0])/2
            AUC_range = np.dot(width,height)
            auc_3d[window]=(AUC_range)
            
            width_PR = TF_list[1:-1,0] - TF_list[:-2,0]
            height_PR = (Precision_list[1:] + Precision_list[:-1])/2
            AP_range = np.dot(width_PR,height_PR)
            ap_3d[window]=(AP_range)

        return tpr_3d, fpr_3d, prec_3d, window_3d, sum(auc_3d)/len(window_3d), sum(ap_3d)/len(window_3d)


def vectorized(tmp, score_sorted, score, l, labels, j, tp, N_pred, L, P, N_labels, TF_list, Precision_list):
    thresholds = score_sorted[tmp]
    pred = score[:, None] >= thresholds
    
    TP = np.sum([np.dot(labels[seg[0]:seg[1]+1], pred[seg[0]:seg[1]+1, :]) for seg in l], axis=0) + tp[j:j+len(tmp)]
    FP = N_pred[j:j+len(tmp)] - TP
    
    existence = np.sum([np.any(np.dot(labels[seg[0]:seg[1]+1], pred[seg[0]:seg[1]+1, :]), axis=0) for seg in L], axis=0)
    existence_ratio = existence / len(L)
    
    P_new = (P + N_labels) / 2
    recall = np.minimum(TP / P_new, 1)
    TPR = recall * existence_ratio
    N_new = len(labels) - P_new
    FPR = FP / N_new
    Precision = TP / N_pred[j:j+len(tmp)]
    
    TF_list[j+1:j+len(tmp)+1] = np.column_stack((TPR, FPR))
    Precision_list[j+1:j+len(tmp)+1] = Precision
    
    return TF_list, Precision_list, j + len(tmp)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import math
from .generics import _sum_wo_nan
"""
In order to shorten the length of the variables,
the general convention in this file is to let:
    - I for a predicted event (start, stop),
    - Is for a list of predicted events,
    - J for a ground truth event,
    - Js for a list of ground truth events.
"""

def interval_length(J = (1,2)):
    """
    Length of an interval
    
    :param J: couple representating the start and stop of an interval, or None
    :return: length of the interval, and 0 for a None interval
    """
    if J is None:
        return(0)
    return(J[1] - J[0])

def sum_interval_lengths(Is = [(1,2),(3,4),(5,6)]):
    """
    Sum of length of the intervals
    
    :param Is: list of intervals represented by starts and stops
    :return: sum of the interval length
    """
    return(sum([interval_length(I) for I in Is]))

def interval_intersection(I = (1, 3), J = (2, 4)): 
    """
    Intersection between two intervals I and J
    I and J should be either empty or represent a positive interval (no point)
    
    :param I: an interval represented by start and stop
    :param J: a second interval of the same form
    :return: an interval representing the start and stop of the intersection (or None if empty)
    """
    if I is None:
        return(None)
    if J is None:
        return(None)
        
    I_inter_J = (max(I[0], J[0]), min(I[1], J[1]))
    if I_inter_J[0] >= I_inter_J[1]:
        return(None)
    else:
        return(I_inter_J)

def interval_subset(I = (1, 3), J = (0, 6)):
    """
    Checks whether I is a subset of J
    
    :param I: an non empty interval represented by start and stop
    :param J: a second non empty interval of the same form
    :return: True if I is a subset of J
    """
    if (I[0] >= J[0]) and (I[1] <= J[1]):
        return True
    else:
        return False

def cut_into_three_func(I, J):
    """
    Cut an interval I into a partition of 3 subsets:
        the elements before J,
        the elements belonging to J,
        and the elements after J
    
    :param I: an interval represented by start and stop, or None for an empty one
    :param J: a non empty interval
    :return: a triplet of three intervals, each represented by either (start, stop) or None
    """
    if I is None:
        return((None, None, None))
    
    I_inter_J = interval_intersection(I, J)
    if I == I_inter_J:
        I_before = None
        I_after = None
    elif I[1] <= J[0]:
        I_before = I
        I_after = None
    elif I[0] >= J[1]:
        I_before = None
        I_after = I
    elif (I[0] <= J[0]) and (I[1] >= J[1]):
        I_before = (I[0], I_inter_J[0])
        I_after = (I_inter_J[1], I[1])
    elif I[0] <= J[0]:
        I_before = (I[0], I_inter_J[0])
        I_after = None
    elif I[1] >= J[1]:
        I_before = None
        I_after = (I_inter_J[1], I[1])
    else:
        raise ValueError('unexpected unconsidered case')
    return(I_before, I_inter_J, I_after)
  
def get_pivot_j(I, J):
    """
    Get the single point of J that is the closest to I, called 'pivot' here,
    with the requirement that I should be outside J
    
    :param I: a non empty interval (start, stop)
    :param J: another non empty interval, with empty intersection with I
    :return: the element j of J that is the closest to I
    """
    if interval_intersection(I, J) is not None:
        raise ValueError('I and J should have a void intersection')

    j_pivot = None # j_pivot is a border of J
    if max(I) <= min(J):
        j_pivot = min(J)
    elif min(I) >= max(J):
        j_pivot = max(J)
    else:
        raise ValueError('I should be outside J')
    return(j_pivot)

def integral_mini_interval(I, J):
    """
    In the specific case where interval I is located outside J,
    integral of distance from x to J over the interval x \in I.
    This is the *integral* i.e. the sum.
    It's not the mean (not divided by the length of I yet)
    
    :param I: a interval (start, stop), or None
    :param J: a non empty interval, with empty intersection with I
    :return: the integral of distances d(x, J) over x \in I
    """
    if I is None:
        return(0)

    j_pivot = get_pivot_j(I, J)
    a = min(I)
    b = max(I)
    return((b-a)*abs((j_pivot - (a+b)/2)))

def integral_interval_distance(I, J):
    """
    For any non empty intervals I, J, compute the
    integral of distance from x to J over the interval x \in I.
    This is the *integral* i.e. the sum. 
    It's not the mean (not divided by the length of I yet)
    The interval I can intersect J or not
    
    :param I: a interval (start, stop), or None
    :param J: a non empty interval
    :return: the integral of distances d(x, J) over x \in I
    """
    # I and J are single intervals (not generic sets)
    # I is a predicted interval in the range of affiliation of J
    
    def f(I_cut):
        return(integral_mini_interval(I_cut, J))
    # If I_middle is fully included into J, it is
    # the distance to J is always 0
    def f0(I_middle):
        return(0)

    cut_into_three = cut_into_three_func(I, J)
    # Distance for now, not the mean:
    # Distance left: Between cut_into_three[0] and the point min(J)
    d_left = f(cut_into_three[0])
    # Distance middle: Between cut_into_three[1] = I inter J, and J
    d_middle = f0(cut_into_three[1])
    # Distance right: Between cut_into_three[2] and the point max(J)
    d_right = f(cut_into_three[2])
    # It's an integral so summable
    return(d_left + d_middle + d_right)

def integral_mini_interval_P_CDFmethod__min_piece(I, J, E):
    """
    Helper of `integral_mini_interval_Pprecision_CDFmethod`
    In the specific case where interval I is located outside J,
    compute the integral $\int_{d_min}^{d_max} \min(m, x) dx$, with:
    - m the smallest distance from J to E,
    - d_min the smallest distance d(x, J) from x \in I to J
    - d_max the largest distance d(x, J) from x \in I to J
    
    :param I: a single predicted interval, a non empty interval (start, stop)
    :param J: ground truth interval, a non empty interval, with empty intersection with I
    :param E: the affiliation/influence zone for J, represented as a couple (start, stop)
    :return: the integral $\int_{d_min}^{d_max} \min(m, x) dx$
    """
    if interval_intersection(I, J) is not None:
        raise ValueError('I and J should have a void intersection')
    if not interval_subset(J, E):
        raise ValueError('J should be included in E')
    if not interval_subset(I, E):
        raise ValueError('I should be included in E')

    e_min = min(E)
    j_min = min(J)
    j_max = max(J)
    e_max = max(E)
    i_min = min(I)
    i_max = max(I)
  
    d_min = max(i_min - j_max, j_min - i_max)
    d_max = max(i_max - j_max, j_min - i_min)
    m = min(j_min - e_min, e_max - j_max)
    A = min(d_max, m)**2 - min(d_min, m)**2
    B = max(d_max, m) - max(d_min, m)
    C = (1/2)*A + m*B
    return(C)

def integral_mini_interval_Pprecision_CDFmethod(I, J, E):
    """
    Integral of the probability of distances over the interval I.
    In the specific case where interval I is located outside J,
    compute the integral $\int_{x \in I} Fbar(dist(x,J)) dx$.
    This is the *integral* i.e. the sum (not the mean)
    
    :param I: a single predicted interval, a non empty interval (start, stop)
    :param J: ground truth interval, a non empty interval, with empty intersection with I
    :param E: the affiliation/influence zone for J, represented as a couple (start, stop)
    :return: the integral $\int_{x \in I} Fbar(dist(x,J)) dx$
    """
    integral_min_piece = integral_mini_interval_P_CDFmethod__min_piece(I, J, E)
  
    e_min = min(E)
    j_min = min(J)
    j_max = max(J)
    e_max = max(E)
    i_min = min(I)
    i_max = max(I)
    d_min = max(i_min - j_max, j_min - i_max)
    d_max = max(i_max - j_max, j_min - i_min)
    integral_linear_piece = (1/2)*(d_max**2 - d_min**2)
    integral_remaining_piece = (j_max - j_min)*(i_max - i_min)
    
    DeltaI = i_max - i_min
    DeltaE = e_max - e_min
    
    output = DeltaI - (1/DeltaE)*(integral_min_piece + integral_linear_piece + integral_remaining_piece)
    return(output)

def integral_interval_probaCDF_precision(I, J, E):
    """
    Integral of the probability of distances over the interval I.
    Compute the integral $\int_{x \in I} Fbar(dist(x,J)) dx$.
    This is the *integral* i.e. the sum (not the mean)
    
    :param I: a single (non empty) predicted interval in the zone of affiliation of J
    :param J: ground truth interval
    :param E: affiliation/influence zone for J
    :return: the integral $\int_{x \in I} Fbar(dist(x,J)) dx$
    """
    # I and J are single intervals (not generic sets)
    def f(I_cut):
        if I_cut is None:
            return(0)
        else:
            return(integral_mini_interval_Pprecision_CDFmethod(I_cut, J, E))
            
    # If I_middle is fully included into J, it is
    # integral of 1 on the interval I_middle, so it's |I_middle|
    def f0(I_middle):
        if I_middle is None:
            return(0)
        else:
            return(max(I_middle) - min(I_middle))
    
    cut_into_three = cut_into_three_func(I, J)
    # Distance for now, not the mean:
    # Distance left: Between cut_into_three[0] and the point min(J)
    d_left = f(cut_into_three[0])
    # Distance middle: Between cut_into_three[1] = I inter J, and J
    d_middle = f0(cut_into_three[1])
    # Distance right: Between cut_into_three[2] and the point max(J)
    d_right = f(cut_into_three[2])
    # It's an integral so summable
    return(d_left + d_middle + d_right)

def cut_J_based_on_mean_func(J, e_mean):
    """
    Helper function for the recall.
    Partition J into two intervals: before and after e_mean
    (e_mean represents the center element of E the zone of affiliation)
    
    :param J: ground truth interval
    :param e_mean: a float number (center value of E)
    :return: a couple partitionning J into (J_before, J_after)
    """
    if J is None:
        J_before = None
        J_after = None
    elif e_mean >= max(J):
        J_before = J
        J_after = None
    elif e_mean <= min(J):
        J_before = None
        J_after = J
    else: # e_mean is across J
        J_before = (min(J), e_mean)
        J_after = (e_mean, max(J))
        
    return((J_before, J_after))

def integral_mini_interval_Precall_CDFmethod(I, J, E):
    """
    Integral of the probability of distances over the interval J.
    In the specific case where interval J is located outside I,
    compute the integral $\int_{y \in J} Fbar_y(dist(y,I)) dy$.
    This is the *integral* i.e. the sum (not the mean)
    
    :param I: a single (non empty) predicted interval
    :param J: ground truth (non empty) interval, with empty intersection with I
    :param E: the affiliation/influence zone for J, represented as a couple (start, stop)
    :return: the integral $\int_{y \in J} Fbar_y(dist(y,I)) dy$
    """
    # The interval J should be located outside I 
    # (so it's either the left piece or the right piece w.r.t I)
    i_pivot = get_pivot_j(J, I)
    e_min = min(E)
    e_max = max(E)
    e_mean = (e_min + e_max) / 2
    
    # If i_pivot is outside E (it's possible), then
    # the distance is worst that any random element within E,
    # so we set the recall to 0
    if i_pivot <= min(E):
        return(0)
    elif i_pivot >= max(E):
        return(0)
    # Otherwise, we have at least i_pivot in E and so d < M so min(d,M)=d
    
    cut_J_based_on_e_mean = cut_J_based_on_mean_func(J, e_mean)
    J_before = cut_J_based_on_e_mean[0]
    J_after = cut_J_based_on_e_mean[1]
  
    iemin_mean = (e_min + i_pivot)/2
    cut_Jbefore_based_on_iemin_mean = cut_J_based_on_mean_func(J_before, iemin_mean)
    J_before_closeE = cut_Jbefore_based_on_iemin_mean[0] # before e_mean and closer to e_min than i_pivot ~ J_before_before
    J_before_closeI = cut_Jbefore_based_on_iemin_mean[1] # before e_mean and closer to i_pivot than e_min ~ J_before_after
    
    iemax_mean = (e_max + i_pivot)/2
    cut_Jafter_based_on_iemax_mean = cut_J_based_on_mean_func(J_after, iemax_mean)
    J_after_closeI = cut_Jafter_based_on_iemax_mean[0] # after e_mean and closer to i_pivot than e_max ~ J_after_before
    J_after_closeE = cut_Jafter_based_on_iemax_mean[1] # after e_mean and closer to e_max than i_pivot ~ J_after_after
    
    if J_before_closeE is not None:
        j_before_before_min = min(J_before_closeE) # == min(J)
        j_before_before_max = max(J_before_closeE)
    else:
        j_before_before_min = math.nan
        j_before_before_max = math.nan
  
    if J_before_closeI is not None:
        j_before_after_min = min(J_before_closeI) # == j_before_before_max if existing
        j_before_after_max = max(J_before_closeI) # == max(J_before)
    else:
        j_before_after_min = math.nan
        j_before_after_max = math.nan
   
    if J_after_closeI is not None:
        j_after_before_min = min(J_after_closeI) # == min(J_after)
        j_after_before_max = max(J_after_closeI) 
    else:
        j_after_before_min = math.nan
        j_after_before_max = math.nan
    
    if J_after_closeE is not None:
        j_after_after_min = min(J_after_closeE) # == j_after_before_max if existing
        j_after_after_max = max(J_after_closeE) # == max(J)
    else:
        j_after_after_min = math.nan
        j_after_after_max = math.nan
  
    # <-- J_before_closeE --> <-- J_before_closeI --> <-- J_after_closeI --> <-- J_after_closeE -->
    # j_bb_min       j_bb_max j_ba_min       j_ba_max j_ab_min      j_ab_max j_aa_min      j_aa_max
    # (with `b` for before and `a` for after in the previous variable names)
    
    #                                          vs e_mean  m = min(t-e_min, e_max-t)  d=|i_pivot-t|   min(d,m)                            \int min(d,m)dt   \int d dt        \int_(min(d,m)+d)dt                                    \int_{t \in J}(min(d,m)+d)dt
    # Case J_before_closeE & i_pivot after J   before     t-e_min                    i_pivot-t       min(i_pivot-t,t-e_min) = t-e_min    t^2/2-e_min*t     i_pivot*t-t^2/2  t^2/2-e_min*t+i_pivot*t-t^2/2 = (i_pivot-e_min)*t      (i_pivot-e_min)*tB - (i_pivot-e_min)*tA = (i_pivot-e_min)*(tB-tA)
    # Case J_before_closeI & i_pivot after J   before     t-e_min                    i_pivot-t       min(i_pivot-t,t-e_min) = i_pivot-t  i_pivot*t-t^2/2   i_pivot*t-t^2/2  i_pivot*t-t^2/2+i_pivot*t-t^2/2 = 2*i_pivot*t-t^2      2*i_pivot*tB-tB^2 - 2*i_pivot*tA + tA^2 = 2*i_pivot*(tB-tA) - (tB^2 - tA^2)
    # Case J_after_closeI & i_pivot after J    after      e_max-t                    i_pivot-t       min(i_pivot-t,e_max-t) = i_pivot-t  i_pivot*t-t^2/2   i_pivot*t-t^2/2  i_pivot*t-t^2/2+i_pivot*t-t^2/2 = 2*i_pivot*t-t^2      2*i_pivot*tB-tB^2 - 2*i_pivot*tA + tA^2 = 2*i_pivot*(tB-tA) - (tB^2 - tA^2)
    # Case J_after_closeE & i_pivot after J    after      e_max-t                    i_pivot-t       min(i_pivot-t,e_max-t) = e_max-t    e_max*t-t^2/2     i_pivot*t-t^2/2  e_max*t-t^2/2+i_pivot*t-t^2/2 = (e_max+i_pivot)*t-t^2  (e_max+i_pivot)*tB-tB^2 - (e_max+i_pivot)*tA + tA^2 = (e_max+i_pivot)*(tB-tA) - (tB^2 - tA^2)
    #
    # Case J_before_closeE & i_pivot before J  before     t-e_min                    t-i_pivot       min(t-i_pivot,t-e_min) = t-e_min    t^2/2-e_min*t     t^2/2-i_pivot*t  t^2/2-e_min*t+t^2/2-i_pivot*t = t^2-(e_min+i_pivot)*t  tB^2-(e_min+i_pivot)*tB - tA^2 + (e_min+i_pivot)*tA = (tB^2 - tA^2) - (e_min+i_pivot)*(tB-tA)
    # Case J_before_closeI & i_pivot before J  before     t-e_min                    t-i_pivot       min(t-i_pivot,t-e_min) = t-i_pivot  t^2/2-i_pivot*t   t^2/2-i_pivot*t  t^2/2-i_pivot*t+t^2/2-i_pivot*t = t^2-2*i_pivot*t      tB^2-2*i_pivot*tB - tA^2 + 2*i_pivot*tA = (tB^2 - tA^2) - 2*i_pivot*(tB-tA)
    # Case J_after_closeI & i_pivot before J   after      e_max-t                    t-i_pivot       min(t-i_pivot,e_max-t) = t-i_pivot  t^2/2-i_pivot*t   t^2/2-i_pivot*t  t^2/2-i_pivot*t+t^2/2-i_pivot*t = t^2-2*i_pivot*t      tB^2-2*i_pivot*tB - tA^2 + 2*i_pivot*tA = (tB^2 - tA^2) - 2*i_pivot*(tB-tA)
    # Case J_after_closeE & i_pivot before J   after      e_max-t                    t-i_pivot       min(t-i_pivot,e_max-t) = e_max-t    e_max*t-t^2/2     t^2/2-i_pivot*t  e_max*t-t^2/2+t^2/2-i_pivot*t = (e_max-i_pivot)*t      (e_max-i_pivot)*tB - (e_max-i_pivot)*tA = (e_max-i_pivot)*(tB-tA)
    
    if i_pivot >= max(J):
        part1_before_closeE = (i_pivot-e_min)*(j_before_before_max - j_before_before_min) # (i_pivot-e_min)*(tB-tA) # j_before_before_max - j_before_before_min
        part2_before_closeI = 2*i_pivot*(j_before_after_max-j_before_after_min) - (j_before_after_max**2 - j_before_after_min**2) # 2*i_pivot*(tB-tA) - (tB^2 - tA^2) # j_before_after_max - j_before_after_min
        part3_after_closeI = 2*i_pivot*(j_after_before_max-j_after_before_min) - (j_after_before_max**2 - j_after_before_min**2) # 2*i_pivot*(tB-tA) - (tB^2 - tA^2) # j_after_before_max - j_after_before_min  
        part4_after_closeE = (e_max+i_pivot)*(j_after_after_max-j_after_after_min) - (j_after_after_max**2 - j_after_after_min**2) # (e_max+i_pivot)*(tB-tA) - (tB^2 - tA^2) # j_after_after_max - j_after_after_min
        out_parts = [part1_before_closeE, part2_before_closeI, part3_after_closeI, part4_after_closeE]
    elif i_pivot <= min(J):
        part1_before_closeE = (j_before_before_max**2 - j_before_before_min**2) - (e_min+i_pivot)*(j_before_before_max-j_before_before_min) # (tB^2 - tA^2) - (e_min+i_pivot)*(tB-tA) # j_before_before_max - j_before_before_min
        part2_before_closeI = (j_before_after_max**2 - j_before_after_min**2) - 2*i_pivot*(j_before_after_max-j_before_after_min) # (tB^2 - tA^2) - 2*i_pivot*(tB-tA) # j_before_after_max - j_before_after_min
        part3_after_closeI = (j_after_before_max**2 - j_after_before_min**2) - 2*i_pivot*(j_after_before_max - j_after_before_min) # (tB^2 - tA^2) - 2*i_pivot*(tB-tA) # j_after_before_max - j_after_before_min
        part4_after_closeE = (e_max-i_pivot)*(j_after_after_max - j_after_after_min) # (e_max-i_pivot)*(tB-tA) # j_after_after_max - j_after_after_min
        out_parts = [part1_before_closeE, part2_before_closeI, part3_after_closeI, part4_after_closeE]
    else:
        raise ValueError('The i_pivot should be outside J')
    
    out_integral_min_dm_plus_d = _sum_wo_nan(out_parts) # integral on all J, i.e. sum of the disjoint parts

    # We have for each point t of J:
    # \bar{F}_{t, recall}(d) = 1 - (1/|E|) * (min(d,m) + d)
    # Since t is a single-point here, and we are in the case where i_pivot is inside E.
    # The integral is then given by:
    # C = \int_{t \in J} \bar{F}_{t, recall}(D(t)) dt
    #   = \int_{t \in J} 1 - (1/|E|) * (min(d,m) + d) dt
    #   = |J| - (1/|E|) * [\int_{t \in J} (min(d,m) + d) dt]
    #   = |J| - (1/|E|) * out_integral_min_dm_plus_d    
    DeltaJ = max(J) - min(J)
    DeltaE = max(E) - min(E)
    C = DeltaJ - (1/DeltaE) * out_integral_min_dm_plus_d
    
    return(C)

def integral_interval_probaCDF_recall(I, J, E):
    """
    Integral of the probability of distances over the interval J.
    Compute the integral $\int_{y \in J} Fbar_y(dist(y,I)) dy$.
    This is the *integral* i.e. the sum (not the mean)

    :param I: a single (non empty) predicted interval
    :param J: ground truth (non empty) interval
    :param E: the affiliation/influence zone for J
    :return: the integral $\int_{y \in J} Fbar_y(dist(y,I)) dy$
    """
    # I and J are single intervals (not generic sets)
    # E is the outside affiliation interval of J (even for recall!)
    # (in particular J \subset E)
    #
    # J is the portion of the ground truth affiliated to I
    # I is a predicted interval (can be outside E possibly since it's recall)
    def f(J_cut):
        if J_cut is None:
            return(0)
        else:
            return integral_mini_interval_Precall_CDFmethod(I, J_cut, E)

    # If J_middle is fully included into I, it is
    # integral of 1 on the interval J_middle, so it's |J_middle|
    def f0(J_middle):
        if J_middle is None:
            return(0)
        else:
            return(max(J_middle) - min(J_middle))
    
    cut_into_three = cut_into_three_func(J, I) # it's J that we cut into 3, depending on the position w.r.t I
    # since we integrate over J this time.
    #
    # Distance for now, not the mean:
    # Distance left: Between cut_into_three[0] and the point min(I)
    d_left = f(cut_into_three[0])
    # Distance middle: Between cut_into_three[1] = J inter I, and I
    d_middle = f0(cut_into_three[1])
    # Distance right: Between cut_into_three[2] and the point max(I)
    d_right = f(cut_into_three[2])
    # It's an integral so summable
    return(d_left + d_middle + d_right)



"""IsolationForest Outlier Detector. Implemented on scikit-learn library.
"""
# Author: Yinchen Wu <Yinchen@uchicago.edu>

from __future__ import division
from __future__ import print_function

from sklearn.ensemble import IsolationForest
from sklearn.utils.validation import check_is_fitted
from sklearn.utils import check_array

from ..utils.utility import invert_order
from .detectorB import DetectorB


class IForest(DetectorB):
    """Wrapper of scikit-learn IsolationForest Class with more functionalities.

    Parameters
    ----------
    n_estimators : int, optional (default=100)
        The number of base estimators in the ensemble.
    max_samples : int, float or string, optional, default="auto"
        The number of samples to draw from X to train each base estimator.

            - If int, then draw `max_samples` samples.
            - If float, then draw `max_samples * X.shape[0]` samples.
            - If "auto", then `max_samples=min(256, n_samples)`.

    contamination : float in (0., 0.5), optional (default=0.1)
        The amount of contamination of the data set, i.e. the proportion
        of outliers in the data set. Used when fitting to define the threshold
        on the decision function.
    max_features : int or float, optional (default=1.0)
        The number of features to draw from X to train each base estimator.

            - If int, then draw `max_features` features.
            - If float, then draw `max_features * X.shape[1]` features.
            - this attribute is useless of a sensor timeseries anomly

    bootstrap : bool, optional (default=False)
        If True, individual trees are fit on random subsets of the training
        data sampled with replacement. If False, sampling without replacement
        is performed.
    n_jobs : integer, optional (default=1)
        The number of jobs to run in parallel for both `fit` and `predict`.
        If -1, then the number of jobs is set to the number of cores.
    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.
    
    Attributes
    ----------
    decision_scores_ : numpy array of shape (n_samples,)
        The outlier scores of the training data.
        The higher, the more abnormal. Outliers tend to have higher
        scores. This value is available once the detector is
        fitted.
    estimators_ : list of DecisionTreeClassifier
        The collection of fitted sub-estimators.
    estimators_samples_ : list of arrays
        The subset of drawn samples (i.e., the in-bag samples) for each base
        estimator.
    max_samples_ : integer
        The actual number of samples
    threshold_ : float
        The threshold is based on ``contamination``. It is the
        ``n_samples * contamination`` most abnormal samples in
        ``decision_scores_``. The threshold is calculated for generating
        binary outlier labels.
    labels_ : int, either 0 or 1
        The binary labels of the training data. 0 stands for inliers
        and 1 for outliers/anomalies. It is generated by applying
        ``threshold_`` on ``decision_scores_``.
    """

    def __init__(self, n_estimators=100,
                 max_samples="auto",
                 contamination=0.1,
                 max_features=1.,
                 bootstrap=False,
                 n_jobs=1,
                 random_state=None,
                 verbose=0):
                
        self.contamination = contamination
        self.n_estimators = n_estimators
        self.max_samples = max_samples
        self.max_features = max_features
        self.bootstrap = bootstrap
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.verbose = verbose
        self.model_name = 'IForest'

    def fit(self, X, y=None):
        """Fit detector. y is ignored in unsupervised methods.
        
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The input samples (time series length). n_features corresponds to the subsequence length.
        y : Ignored
            Not used, present for API consistency by convention.
        
        Returns
        -------
        self : object
            Fitted estimator.
        """
        # validate inputs X and y (optional)
        try: 
            X = check_array(X)
        except:
            X = X.reshape(-1, 1)
        
        X = check_array(X)
            
        self.detector_ = IsolationForest(n_estimators=self.n_estimators,
                                            max_samples=self.max_samples,
                                            contamination=self.contamination,
                                            max_features=self.max_features,
                                            bootstrap=self.bootstrap,
                                            n_jobs=self.n_jobs,
                                            random_state=self.random_state,
                                            verbose=self.verbose)


        self.detector_.fit(X=X, y=None, sample_weight=None)

        # invert decision_scores_. Outliers comes with higher outlier scores.
        self.decision_scores_ = -self.detector_.score_samples(X)
        return self

    def decision_function(self, X):
        """Predict raw anomaly score of X using the fitted detector.
        The anomaly score of an input sample is computed based on different
        detector algorithms. For consistency, outliers are assigned with
        larger anomaly scores.
        
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The training input samples. Sparse matrices are accepted only
            if they are supported by the base estimator.
        
        Returns
        -------
        anomaly_scores : numpy array of shape (n_samples,)
            The anomaly score of the input samples.
        """
        check_is_fitted(self, ['decision_scores_', 'threshold_', 'labels_'])
        # invert outlier scores. Outliers comes with higher outlier scores
        return invert_order(self.detector_.decision_function(X))



    def estimators_(self):
        """The collection of fitted sub-estimators.
        Decorator for scikit-learn Isolation Forest attributes.

        :meta private:
        """
        return self.detector_.estimators_


    def estimators_samples_(self):
        """The subset of drawn samples (i.e., the in-bag samples) for
        each base estimator.
        Decorator for scikit-learn Isolation Forest attributes.

        :meta private:
        """
        return self.detector_.estimators_samples_

    def max_samples_(self):
        """The actual number of samples.
        Decorator for scikit-learn Isolation Forest attributes.

        :meta private:
        """
        return self.detector_.max_samples_



from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Conv1D, MaxPooling1D, Flatten, Dropout
from tensorflow.keras.callbacks import EarlyStopping


import numpy as np
from sklearn.preprocessing import MinMaxScaler

class cnn:
    """
    Implementation of CNN
    
    Parameters
    ----------
    slidingwindow : int
        Subsequence length to analyze.
    predict_time_steps : int, (default=1)
        The length of the subsequence to predict.
    epochs : int, (default=10)
        Number of epochs for the training phase
    patience : int, (default=10)
        Number of epoch to wait before early stopping during training

    Attributes
    ----------
    decision_scores_ : numpy array of shape (n_samples - subsequence_length,)
        The anomaly score.
        The higher, the more abnormal. Anomalies tend to have higher
        scores. This value is available once decision_function is called.
    """
    def __init__(self, slidingwindow = 100, predict_time_steps=1, epochs = 10, patience = 10, verbose=0):
        self.slidingwindow = slidingwindow
        self.predict_time_steps = predict_time_steps
        self.epochs = epochs
        self.patience = patience
        self.verbose = verbose
        self.model_name = 'cnn'
        
    def fit(self, X_clean, X_dirty, ratio = 0.15):
        """Fit detector.
        
        Parameters
        ----------
        X_clean : numpy array of shape (n_samples, )
            The input training samples.
        X_dirty : numpy array of shape (n_samples, )
            The input testing samples.
        ratio : flaot, ([0,1])
            The ratio for the train validation split
        
        Returns
        -------
        self : object
            Fitted estimator.
        """

        slidingwindow = self.slidingwindow
        predict_time_steps = self.predict_time_steps
        self.n_test_ = len(X_dirty)

        X_train, Y_train = self.create_dataset(X_clean, slidingwindow, predict_time_steps)
        X_test, Y_test = self.create_dataset(X_dirty, slidingwindow, predict_time_steps)
        
        X_train = X_train.reshape((X_train.shape[0], X_train.shape[1], 1))
        X_test = X_test.reshape((X_test.shape[0], X_test.shape[1], 1))
        
        model = Sequential()
        model.add(Conv1D(filters=8,
                         kernel_size=2,
                         strides=1,
                         padding='same',
                         activation='relu',
                         input_shape=(slidingwindow, 1)))
        model.add(MaxPooling1D(pool_size=2)) 
        model.add(Conv1D(filters=16,
                         kernel_size=2,
                         strides=1,
                         padding='valid',
                         activation='relu'))
        model.add(MaxPooling1D(pool_size=2))
        model.add(Conv1D(filters=32,
                         kernel_size=2,
                         strides=1,
                         padding='valid',
                         activation='relu'))
        model.add(MaxPooling1D(pool_size=2))
        model.add(Flatten())
        model.add(Dense(units=64, activation='relu'))  
        model.add(Dropout(rate=0.2))
        model.add(Dense(units=predict_time_steps))
        
        model.compile(loss='mse', optimizer='adam')
        
        es = EarlyStopping(monitor='val_loss', mode='min', verbose=self.verbose, patience=self.patience)
        
        model.fit(X_train,Y_train,validation_split=ratio,
                  epochs=self.epochs,batch_size=64,verbose=self.verbose, callbacks=[es])
        
        prediction = model.predict(X_test)

        self.Y_test = Y_test
        self.prediction = prediction
        self.estimator = model
        self.n_initial = X_train.shape[0]
        
        return self
        
    def create_dataset(self, X, slidingwindow, predict_time_steps=1):
        Xs, ys = [], []
        for i in range(len(X) - slidingwindow - predict_time_steps+1):
            tmp = X[i : i + slidingwindow + predict_time_steps]
            tmp= MinMaxScaler(feature_range=(0,1)).fit_transform(tmp.reshape(-1,1)).ravel()
            x = tmp[:slidingwindow]
            y = tmp[slidingwindow:]
            
            Xs.append(x)
            ys.append(y)
        return np.array(Xs), np.array(ys)
    

    
    def decision_function(self, measure = None):
        """Derive the decision score based on the given distance measure
        
        Parameters
        ----------
        measure : object
            object for given distance measure with methods to derive the score
        
        Returns
        -------
        self : object
            Fitted estimator.
        """
        
        Y_test = self.Y_test

        score = []
        prediction = self.prediction

        for i in range(prediction.shape[0]):
            score.append(measure.measure(Y_test[i], prediction[i], 0))
        
        score = np.array(score)
        decision_scores_ = np.zeros(self.n_test_)
        
        decision_scores_[self.slidingwindow : (self.n_test_-self.predict_time_steps+1)]=score
        decision_scores_[: self.slidingwindow] = score[0]
        decision_scores_[self.n_test_-self.predict_time_steps+1:]=score[-1]
        
        self.decision_scores_ = decision_scores_
        return self


from statsmodels.tsa.stattools import acf
from scipy.signal import argrelextrema
import numpy as np

import matplotlib.patches as mpatches 
import matplotlib.pyplot as plt

from ..vus.utils.metrics import metricor

def plotFig(data, label, score, slidingWindow, fileName, modelName, plotRange=None):
    grader = metricor()
    
    range_anomaly = grader.range_convers_new(label)
    
    max_length = len(score)

    if plotRange==None:
        plotRange = [0,max_length]
    
    fig3 = plt.figure(figsize=(12, 10), constrained_layout=True)
    gs = fig3.add_gridspec(3, 4)
    
    
    f3_ax1 = fig3.add_subplot(gs[0, :-1])
    plt.tick_params(labelbottom=False)

    plt.plot(data[:max_length],'k')
    for r in range_anomaly:
        if r[0]==r[1]:
            plt.plot(r[0],data[r[0]],'r.')
        else:
            plt.plot(range(r[0],r[1]+1),data[range(r[0],r[1]+1)],'r')
    
    plt.xlim(plotRange)
    

    f3_ax2 = fig3.add_subplot(gs[1, :-1])
    
    plt.plot(score[:max_length])
    plt.hlines(np.mean(score)+3*np.std(score),0,max_length,linestyles='--',color='red')
    plt.ylabel('score')
    
    plt.xlim(plotRange)
    
    
    #plot the data
    f3_ax3 = fig3.add_subplot(gs[2, :-1])
    index = ( label + 2*(score > (np.mean(score)+3*np.std(score))))
    cf = lambda x: 'k' if x==0 else ('r' if x == 1 else ('g' if x == 2 else 'b') )
    cf = np.vectorize(cf)
    
    color = cf(index[:max_length])
    black_patch = mpatches.Patch(color = 'black', label = 'TN')
    red_patch = mpatches.Patch(color = 'red', label = 'FN')
    green_patch = mpatches.Patch(color = 'green', label = 'FP')
    blue_patch = mpatches.Patch(color = 'blue', label = 'TP')
    plt.scatter(np.arange(max_length), data[:max_length], c=color, marker='.')
    plt.legend(handles = [black_patch, red_patch, green_patch, blue_patch], loc= 'best')
    plt.xlim(plotRange)
    
        


from setuptools import setup, find_packages

__version__ = "0.0.3"

CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: Apache Software License",
    "Natural Language :: English",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
    "Topic :: Scientific/Engineering :: Information Analysis",
    "Topic :: Scientific/Engineering :: Mathematics",
    "Topic :: Software Development :: Libraries :: Python Modules",
]

setup(
    name="TSB_UAD",
    version=__version__,
    description="Time Series Anomaly Detection Benchmark",
    classifiers=CLASSIFIERS,
    author="The DATUM Lab",
    author_email="john@paparrizos.org",
    packages=find_packages(),
    zip_safe=True,
    license="Apache-2.0 license",
    url="https://github.com/TheDatumOrg/TSB-UAD",
    entry_points={},
    install_requires=[
        'numpy>=1.24.3',
        'matplotlib>=3.7.5',
        'pandas>=2.0.3',
        'arch>=5.3.1',
        'tsfresh>=0.20.2',
        'hurst>=0.0.5',
        'tslearn>=0.6.3',
        'cython>=3.0.10',
        'scikit-learn>=1.3.2',
        'stumpy>=1.12.0',
        'tensorflow>=2.13.0',
        'networkx>=3.1',
        ]
)
